from sapiopylib.rest.ELNService import ElnManager
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import CustomReportCriteria, AbstractReportTerm, RawReportTerm
from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType
from sapiopylib.rest.pojo.eln.ElnExperiment import ElnExperiment, ElnExperimentQueryCriteria
from sapiopylib.rest.pojo.eln.SapioELNEnums import ElnExperimentStatus, ElnBaseDataType
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.customreport.custom_report_builder import CustomReportBuilder
from sapiopycommons.customreport.term_builder import TermBuilder
from sapiopycommons.datatype.pseudo_data_types import EnbEntryOptionsPseudoDef, NotebookExperimentOptionPseudoDef, \
    NotebookExperimentPseudoDef, ExperimentEntryRecordPseudoDef, EnbEntryPseudoDef
from sapiopycommons.general.aliases import SapioRecord, UserIdentifier, AliasUtil, FieldValue, \
    ExperimentEntryIdentifier, ExperimentIdentifier
from sapiopycommons.general.custom_report_util import CustomReportUtil
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.recordmodel.record_handler import RecordHandler


# FR-47234: Create an ExperimentReportCriteria class for finer experiment searching
# and create/update ExperimentReportUtil functions to use it.
class ExperimentReportCriteria:
    """
    Experiment report criteria is used to restrict the results of searches for experiments.
    """
    notebook_ids: list[int] | None
    not_notebook_ids: list[int] | None
    names: list[str] | None
    not_names: list[str] | None
    created_by: list[str] | None
    not_created_by: list[str] | None
    last_modified_by: list[str] | None
    not_last_modified_by: list[str] | None
    owners: list[str] | None
    not_owners: list[str] | None
    statuses: list[str] | None
    not_statuses: list[str] | None
    templates: str | None
    not_templates: str | None
    is_from_template: bool | None
    is_from_protocol_template: bool | None
    is_modifiable: bool | None
    is_active: bool | None
    created_before: int | None
    created_after: int | None
    due_before: int | None
    due_after: int | None
    last_modified_before: int | None
    last_modified_after: int | None

    def __init__(self, *,
                 notebook_ids: list[ExperimentIdentifier] | None = None,
                 not_notebook_ids: list[ExperimentIdentifier] | None = None,
                 names: list[str] | None = None,
                 not_names: list[str] | None = None,
                 created_by: list[str] | None = None,
                 not_created_by: list[str] | None = None,
                 last_modified_by: list[str] | None = None,
                 not_last_modified_by: list[str] | None = None,
                 owners: list[str] | None = None,
                 not_owners: list[str] | None = None,
                 statuses: list[ElnExperimentStatus | str] | None = None,
                 not_statuses: list[ElnExperimentStatus | str] | None = None,
                 templates: list[str] | None = None,
                 not_templates: list[str] | None = None,
                 is_from_template: bool | None = None,
                 is_from_protocol_template: bool | None = None,
                 is_modifiable: bool | None = None,
                 is_active: bool | None = None,
                 created_after: int | None = None,
                 created_before: int | None = None,
                 due_after: int | None = None,
                 due_before: int | None = None,
                 last_modified_after: int | None = None,
                 last_modified_before: int | None = None):
        """
        Restrict searches using the following criteria.

        :param notebook_ids: The allowed notebook ID(s) of the experiment.
        :param not_notebook_ids: The disallowed notebook ID(s) of the experiment.
        :param names: The allowed name(s) of the experiment.
        :param not_names: The disallowed name(s) of the experiment.
        :param created_by: The allowed username(s) of the user who created the experiment.
        :param not_created_by: The disallowed username(s) of the user who created the experiment.
        :param last_modified_by: The allowed username(s) of the user who last modified the experiment.
        :param not_last_modified_by: The disallowed username(s) of the user who last modified the experiment.
        :param owners: The allowed username(s) of the user who owns the experiment.
        :param not_owners: The disallowed username(s) of the user who owns the experiment.
        :param statuses: The allowed status(es) of the experiment.
        :param not_statuses: The disallowed status(es) of the experiment.
        :param templates: The allowed template name(s) that the experiment was created from.
        :param not_templates: The disallowed template name(s) that the experiment was created from.
        :param is_from_template: Whether the experiment was created from a template.
        :param is_from_protocol_template: Whether the experiment was created from a protocol template.
        :param is_modifiable: Whether the experiment is modifiable.
        :param is_active: Whether the experiment is from an active template.
        :param created_after: A timestamp after which the experiment was created.
        :param created_before: A timestamp before which the experiment was created.
        :param due_after: A timestamp after which the experiment's approval is due.
        :param due_before: A timestamp before which the experiment's approval is due.
        :param last_modified_after: A timestamp after which the experiment last modified.
        :param last_modified_before: A timestamp before which the experiment last modified.
        """
        self.notebook_ids = notebook_ids
        self.not_notebook_ids = not_notebook_ids
        self.names = names
        self.not_names = not_names
        self.created_by = created_by
        self.not_created_by = not_created_by
        self.last_modified_by = last_modified_by
        self.not_last_modified_by = not_last_modified_by
        self.owners = owners
        self.not_owners = not_owners
        self.statuses = statuses
        self.not_statuses = not_statuses
        self.templates = templates
        self.not_templates = not_templates

        self.is_from_template = is_from_template
        self.is_from_protocol_template = is_from_protocol_template
        self.is_modifiable = is_modifiable
        self.is_active = is_active

        self.created_before = created_before
        self.created_after = created_after
        self.due_before = due_before
        self.due_after = due_after
        self.last_modified_before = last_modified_before
        self.last_modified_after = last_modified_after

        if self.notebook_ids is not None:
            self.notebook_ids = AliasUtil.to_notebook_ids(self.notebook_ids)
        if self.not_notebook_ids is not None:
            self.not_notebook_ids = AliasUtil.to_notebook_ids(self.not_notebook_ids)
        if self.statuses is not None:
            self.statuses = [x.description if isinstance(x, ElnExperimentStatus) else x for x in self.statuses]
        if self.not_statuses is not None:
            self.not_statuses = [x.description if isinstance(x, ElnExperimentStatus) else x for x in self.not_statuses]


# FR-46908 - Provide a utility class that holds experiment related custom reports e.g. getting all the experiments
# that given records were used in or getting all records of a datatype used in given experiments.
class ExperimentReportUtil:
    @staticmethod
    def map_records_to_experiment_ids(context: UserIdentifier, records: list[SapioRecord]) \
            -> dict[SapioRecord, list[int]]:
        """
        Return a dictionary mapping each record to a list of ids of experiments that they were used in.
        If a record wasn't used in any experiments then it will be mapped to an empty list.

        :param context: The current webhook context or a user object to send requests from.
        :param records: A list of records of the same data type.
        :return: A dictionary mapping each record to a list of ids of each experiment it was used in.
        """
        if not records:
            return {}

        user: SapioUser = AliasUtil.to_sapio_user(context)
        data_type_name: str = AliasUtil.to_singular_data_type_name(records)

        record_ids: list[int] = AliasUtil.to_record_ids(records)
        rows = ExperimentReportUtil.__get_record_experiment_relation_rows(user, data_type_name, record_ids=record_ids)

        id_to_record: dict[int, SapioRecord] = RecordHandler.map_by_id(records)
        record_to_exps: dict[SapioRecord, set[int]] = {record: set() for record in records}
        for row in rows:
            record_id: int = row["RecordId"]
            exp_id: int = row[NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name]
            record = id_to_record[record_id]
            record_to_exps[record].add(exp_id)

        return {record: list(exps) for record, exps in record_to_exps.items()}

    @staticmethod
    def map_eln_records_to_experiment_ids(context: UserIdentifier, records: list[SapioRecord]) \
            -> dict[SapioRecord, int]:
        """
        Return a dictionary mapping each ELN record to the ID of the experiments that each were used in.

        :param context: The current webhook context or a user object to send requests from.
        :param records: A list of ELN data type records. They are permitted to be of mixed data type names.
        :return: A dictionary mapping each record to the ID of the experiment it was used in.
        """
        if not records:
            return {}

        data_types: set[str] = AliasUtil.to_data_type_names(records, True, False)
        for data_type in data_types:
            if not ElnBaseDataType.is_eln_type(data_type):
                raise SapioException(f"Data type {data_type} is not an ELN data type.")

        dt_to_record: dict[str, list[SapioRecord]] = {}
        for record in records:
            dt_to_record.setdefault(AliasUtil.to_data_type_name(record, False), []).append(record)

        report_builder = CustomReportBuilder(EnbEntryPseudoDef.DATA_TYPE_NAME)
        tb = report_builder.get_term_builder()
        report_builder.add_column(EnbEntryPseudoDef.DATA_TYPE_NAME__FIELD_NAME)
        report_builder.add_column(EnbEntryPseudoDef.EXPERIMENT_ID__FIELD_NAME)
        report_builder.set_root_term(tb.is_term(EnbEntryPseudoDef.DATA_TYPE_NAME__FIELD_NAME, data_types))
        criteria = report_builder.build_report_criteria()

        ret_val: dict[SapioRecord, int] = {}
        rows: list[dict[str, FieldValue]] = CustomReportUtil.run_custom_report(context, criteria)
        for row in rows:
            dt: str = row[EnbEntryPseudoDef.DATA_TYPE_NAME__FIELD_NAME.field_name]
            exp_id: int = row[EnbEntryPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name]
            for record in dt_to_record[dt]:
                ret_val[record] = exp_id
        return ret_val

    @staticmethod
    def map_experiments_to_records_of_type(context: UserIdentifier, exp_ids: list[ExperimentIdentifier],
                                           wrapper_type: type[WrappedType]) -> dict[int, list[WrappedType]]:
        """
        Return a dictionary mapping each experiment id to a list of records of the given type that were used in each
        experiment. If an experiment didn't use any records of the given type then it will be mapped to an empty list.

        :param context: The current webhook context or a user object to send requests from.
        :param exp_ids: A list of experiment identifiers.
        :param wrapper_type: The record model wrapper to use, corresponds to which data type we will query for.
        :return: A dictionary mapping each experiment id to a list of records of the given type that were used in that
            experiment.
        """
        if not exp_ids:
            return {}

        user = AliasUtil.to_sapio_user(context)
        record_handler = RecordHandler(user)
        data_type_name: str = wrapper_type.get_wrapper_data_type_name()

        exp_ids: list[int] = AliasUtil.to_notebook_ids(exp_ids)
        rows = ExperimentReportUtil.__get_record_experiment_relation_rows(user, data_type_name, exp_ids=exp_ids)
        record_ids: set[int] = {row["RecordId"] for row in rows}
        records = record_handler.query_models_by_id(wrapper_type, record_ids)

        id_to_record: dict[int, WrappedType] = RecordHandler.map_by_id(records)
        exp_to_records: dict[int, set[SapioRecord]] = {exp: set() for exp in exp_ids}
        for row in rows:
            record_id: int = row["RecordId"]
            exp_id: int = row[NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name]
            record = id_to_record[record_id]
            exp_to_records[exp_id].add(record)

        return {exp: list(records) for exp, records in exp_to_records.items()}

    @staticmethod
    def get_experiment_options(context: UserIdentifier, experiments: list[ExperimentIdentifier]) \
            -> dict[int, dict[str, str]]:
        """
        Run a custom report to retrieve the experiment options for all the provided experiments. Effectively a batched
        version of the get_notebook_experiment_options function of ElnManager.

        :param context: The current webhook context or a user object to send requests from.
        :param experiments: The experiment identifiers to retrieve the experiment options for.
        :return: A dictionary mapping the notebook experiment ID to the options for that experiment.
        """
        exp_ids: list[int] = AliasUtil.to_notebook_ids(experiments)

        report_builder = CustomReportBuilder(NotebookExperimentOptionPseudoDef.DATA_TYPE_NAME)
        tb = report_builder.get_term_builder()
        root = tb.is_term(NotebookExperimentOptionPseudoDef.EXPERIMENT_ID__FIELD_NAME, exp_ids)
        report_builder.set_root_term(root)
        report_builder.add_column(NotebookExperimentOptionPseudoDef.EXPERIMENT_ID__FIELD_NAME)
        report_builder.add_column(NotebookExperimentOptionPseudoDef.OPTION_KEY__FIELD_NAME)
        report_builder.add_column(NotebookExperimentOptionPseudoDef.OPTION_VALUE__FIELD_NAME)
        report = report_builder.build_report_criteria()

        # Ensure that each experiment appears in the dictionary, even if it has no experiment options.
        options: dict[int, dict[str, str]] = {x: {} for x in exp_ids}
        results: list[dict[str, FieldValue]] = CustomReportUtil.run_custom_report(context, report)
        for row in results:
            exp_id: int = row[NotebookExperimentOptionPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name]
            key: str = row[NotebookExperimentOptionPseudoDef.OPTION_KEY__FIELD_NAME.field_name]
            value: str = row[NotebookExperimentOptionPseudoDef.OPTION_VALUE__FIELD_NAME.field_name]
            options[exp_id][key] = value
        return options

    @staticmethod
    def get_experiment_entry_options(context: UserIdentifier, entries: list[ExperimentEntryIdentifier]) \
            -> dict[int, dict[str, str]]:
        """
        Run a custom report to retrieve the entry options for all the provided entries. Effectively a batched
        version of the get_experiment_entry_options function of ElnManager.

        :param context: The current webhook context or a user object to send requests from.
        :param entries: The experiment entry identifiers to retrieve the entry options for.
        :return: A dictionary mapping the entry ID to the options for that entry.
        """
        entries: list[int] = AliasUtil.to_entry_ids(entries)
        report_builder = CustomReportBuilder(EnbEntryOptionsPseudoDef.DATA_TYPE_NAME)
        tb = report_builder.get_term_builder()
        root = tb.is_term(EnbEntryOptionsPseudoDef.ENTRY_ID__FIELD_NAME, entries)
        report_builder.set_root_term(root)
        report_builder.add_column(EnbEntryOptionsPseudoDef.ENTRY_ID__FIELD_NAME)
        report_builder.add_column(EnbEntryOptionsPseudoDef.ENTRY_OPTION_KEY__FIELD_NAME)
        report_builder.add_column(EnbEntryOptionsPseudoDef.ENTRY_OPTION_VALUE__FIELD_NAME)
        report = report_builder.build_report_criteria()

        # Ensure that each entry appears in the dictionary, even if it has no entry options.
        options: dict[int, dict[str, str]] = {x: {} for x in entries}
        results: list[dict[str, FieldValue]] = CustomReportUtil.run_custom_report(context, report)
        for row in results:
            entry_id: int = row[EnbEntryOptionsPseudoDef.ENTRY_ID__FIELD_NAME.field_name]
            key: str = row[EnbEntryOptionsPseudoDef.ENTRY_OPTION_KEY__FIELD_NAME.field_name]
            value: str = row[EnbEntryOptionsPseudoDef.ENTRY_OPTION_VALUE__FIELD_NAME.field_name]
            options[entry_id][key] = value
        return options

    @staticmethod
    def get_template_names_for_experiments(context: UserIdentifier, experiments: list[ExperimentIdentifier]) \
            -> dict[int, str]:
        """
        Run a custom report to retrieve the template names for all the provided experiments.

        :param context: The current webhook context or a user object to send requests from.
        :param experiments: The experiment identifiers to retrieve the template names for.
        :return: A dictionary mapping the notebook experiment ID to the template name that the experiment was created from.
        """
        exp_ids: list[int] = AliasUtil.to_notebook_ids(experiments)

        report_builder = CustomReportBuilder(NotebookExperimentPseudoDef.DATA_TYPE_NAME)
        tb = report_builder.get_term_builder()
        root = tb.is_term(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME, exp_ids)
        report_builder.add_join(tb.compare_is_term("ELNExperiment",
                                                   "RecordId",
                                                   NotebookExperimentPseudoDef.DATA_TYPE_NAME,
                                                   NotebookExperimentPseudoDef.EXPERIMENT_RECORD_ID__FIELD_NAME))
        report_builder.set_root_term(root)
        report_builder.add_column(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME)
        report_builder.add_column("TemplateExperimentName", FieldType.STRING, data_type="ELNExperiment")
        report = report_builder.build_report_criteria()

        ret_val: dict[int, str] = {}
        results: list[dict[str, FieldValue]] = CustomReportUtil.run_custom_report(context, report)
        for row in results:
            exp_id: int = row[NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name]
            name: str = row["TemplateExperimentName"]
            ret_val[exp_id] = name
        return ret_val

    @staticmethod
    def get_experiments_for_criteria(context: UserIdentifier, criteria: ExperimentReportCriteria) -> list[ElnExperiment]:
        """
        Run a custom report that retrieves every experiment in the system for the given search criteria.

        :param context: The current webhook context or a user object to send requests from.
        :param criteria: The search criteria to query for experiments with.
        :return: A list of every experiment in the system that matches the search criteria.
        """
        report: CustomReportCriteria = ExperimentReportUtil.build_experiment_id_report(criteria)
        return ExperimentReportUtil.get_experiments_for_report(context, report)

    @staticmethod
    def get_experiments_by_name(context: UserIdentifier, name: str,
                                criteria: ExperimentReportCriteria = ExperimentReportCriteria()) -> list[ElnExperiment]:
        """
        Run a custom report that retrieves every experiment in the system with a given name.

        :param context: The current webhook context or a user object to send requests from.
        :param name: The name of the experiment to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A list of every experiment in the system with a name that matches the input
            that matches the given criteria.
        """
        return ExperimentReportUtil.get_experiments_by_names(context, [name], criteria)[name]

    @staticmethod
    def get_experiments_by_names(context: UserIdentifier, names: list[str],
                                 criteria: ExperimentReportCriteria = ExperimentReportCriteria()) -> dict[str, list[ElnExperiment]]:
        """
        Run a custom report that retrieves every experiment in the system with a name from a list of names.

        :param context: The current webhook context or a user object to send requests from.
        :param names: The names of the experiment to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A dictionary mapping the experiment name to a list of every experiment in the system with that name
            that matches the given criteria.
        """
        criteria.names = names
        experiments: list[ElnExperiment] = ExperimentReportUtil.get_experiments_for_criteria(context, criteria)

        # Ensure that each name appears in the dictionary, even if it has no experiments.
        ret_val: dict[str, list[ElnExperiment]] = {x: [] for x in names}
        for experiment in experiments:
            ret_val.get(experiment.notebook_experiment_name).append(experiment)
        return ret_val

    @staticmethod
    def get_experiments_by_owner(context: UserIdentifier, owner: str,
                                 criteria: ExperimentReportCriteria = ExperimentReportCriteria()) -> list[ElnExperiment]:
        """
        Run a custom report that retrieves every experiment in the system with a given owner.

        :param context: The current webhook context or a user object to send requests from.
        :param owner: The username of the owner of the experiments to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A list of every experiment in the system with the owner that matches the input
            that matches the given criteria.
        """
        return ExperimentReportUtil.get_experiments_by_owners(context, [owner], criteria)[owner]

    @staticmethod
    def get_experiments_by_owners(context: UserIdentifier, owners: list[str],
                                  criteria: ExperimentReportCriteria = ExperimentReportCriteria) \
            -> dict[str, list[ElnExperiment]]:
        """
        Run a custom report that retrieves every experiment in the system with a given list of owners.

        :param context: The current webhook context or a user object to send requests from.
        :param owners: The usernames of the owner of the experiments to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A dictionary mapping the owner username to a list of every experiment in the system from that owner
            that matches the given criteria.
        """
        criteria.owners = owners
        experiments: list[ElnExperiment] = ExperimentReportUtil.get_experiments_for_criteria(context, criteria)

        # Ensure that each name appears in the dictionary, even if it has no experiments.
        ret_val: dict[str, list[ElnExperiment]] = {x: [] for x in owners}
        for experiment in experiments:
            ret_val.get(experiment.owner).append(experiment)
        return ret_val

    @staticmethod
    def get_experiments_by_creator(context: UserIdentifier, created_by: str,
                                   criteria: ExperimentReportCriteria = ExperimentReportCriteria()) \
            -> list[ElnExperiment]:
        """
        Run a custom report that retrieves every experiment in the system with a given creator.

        :param context: The current webhook context or a user object to send requests from.
        :param created_by: The username of the creator of the experiments to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A list of every experiment in the system with the creator that matches the input
            that matches the given criteria.
        """
        return ExperimentReportUtil.get_experiments_by_creators(context, [created_by], criteria=criteria)[created_by]

    @staticmethod
    def get_experiments_by_creators(context: UserIdentifier, created_by: list[str],
                                    criteria: ExperimentReportCriteria = ExperimentReportCriteria()) \
            -> dict[str, list[ElnExperiment]]:
        """
        Run a custom report that retrieves every experiment in the system with a given list of creators.

        :param context: The current webhook context or a user object to send requests from.
        :param created_by: The usernames of the creator of the experiments to query for.
        :param criteria: Additional search criteria to filter the results.
        :return: A dictionary mapping the owner username to a list of every experiment in the system from that creator
            that matches the given criteria.
        """
        criteria.created_by = created_by
        experiments: list[ElnExperiment] = ExperimentReportUtil.get_experiments_for_criteria(context, criteria)

        # Ensure that each name appears in the dictionary, even if it has no experiments.
        ret_val: dict[str, list[ElnExperiment]] = {x: [] for x in created_by}
        for experiment in experiments:
            ret_val.get(experiment.created_by).append(experiment)
        return ret_val

    @staticmethod
    def build_experiment_id_report(criteria: ExperimentReportCriteria = ExperimentReportCriteria()) \
            -> CustomReportCriteria:
        """
        Construct a custom report using the provided ExperimentReportCriteria.

        :param criteria: The criteria to construct a custom report from.
        :return: A custom report that can be used to search for experiment IDs that match the given criteria.
        """
        dt: str = NotebookExperimentPseudoDef.DATA_TYPE_NAME
        report_builder = CustomReportBuilder(dt)
        tb = report_builder.get_term_builder()
        report_builder.add_column(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME)

        root: AbstractReportTerm | None = None
        if criteria.notebook_ids is not None:
            root = tb.is_term(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME, criteria.notebook_ids)
        if criteria.not_notebook_ids is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME, criteria.not_notebook_ids)
            if root is not None:
                root = tb.and_terms(root, term)
        if root is None:
            root = ExperimentReportUtil.all_notebook_experiments_term()

        if criteria.names is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.EXPERIMENT_NAME__FIELD_NAME, criteria.names)
            root = tb.and_terms(root, term)
        if criteria.not_names is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.EXPERIMENT_NAME__FIELD_NAME, criteria.not_names)
            root = tb.and_terms(root, term)

        if criteria.created_by is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.CREATED_BY__FIELD_NAME, criteria.created_by)
            root = tb.and_terms(root, term)
        if criteria.not_created_by is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.CREATED_BY__FIELD_NAME, criteria.not_created_by)
            root = tb.and_terms(root, term)

        if criteria.last_modified_by is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.LAST_MODIFIED_BY__FIELD_NAME, criteria.last_modified_by)
            root = tb.and_terms(root, term)
        if criteria.not_last_modified_by is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.LAST_MODIFIED_BY__FIELD_NAME, criteria.not_last_modified_by)
            root = tb.and_terms(root, term)

        if criteria.owners is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.EXPERIMENT_OWNER__FIELD_NAME, criteria.owners)
            root = tb.and_terms(root, term)
        if criteria.not_owners is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.EXPERIMENT_OWNER__FIELD_NAME, criteria.not_owners)
            root = tb.and_terms(root, term)

        if criteria.statuses is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.STATUS__FIELD_NAME, criteria.statuses)
            root = tb.and_terms(root, term)
        if criteria.not_statuses is not None:
            term = tb.not_term(NotebookExperimentPseudoDef.STATUS__FIELD_NAME, criteria.not_statuses)
            root = tb.and_terms(root, term)

        # For the template name term, we need to join on the experiment record.
        if criteria.templates is not None or criteria.not_templates is not None:
            join = tb.compare_is_term("ELNExperiment",
                                      "RecordId",
                                      NotebookExperimentPseudoDef.DATA_TYPE_NAME,
                                      NotebookExperimentPseudoDef.EXPERIMENT_RECORD_ID__FIELD_NAME)
            report_builder.add_join(join)
        if criteria.templates is not None:
            term = tb.is_term("TemplateExperimentName", criteria.templates, data_type="ELNExperiment")
            root = tb.and_terms(root, term)
        if criteria.not_templates is not None:
            term = tb.not_term("TemplateExperimentName", criteria.not_templates, data_type="ELNExperiment")
            root = tb.and_terms(root, term)

        if criteria.is_from_template is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.IS_TEMPLATE__FIELD_NAME, criteria.is_from_template)
            root = tb.and_terms(root, term)
        if criteria.is_from_protocol_template is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.IS_PROTOCOL_TEMPLATE__FIELD_NAME, criteria.is_from_protocol_template)
            root = tb.and_terms(root, term)
        if criteria.is_modifiable is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.IS_MODIFIABLE__FIELD_NAME, criteria.is_modifiable)
            root = tb.and_terms(root, term)
        if criteria.is_active is not None:
            term = tb.is_term(NotebookExperimentPseudoDef.IS_ACTIVE__FIELD_NAME, criteria.is_active)
            root = tb.and_terms(root, term)

        if criteria.created_after is not None:
            term = tb.gte_term(NotebookExperimentPseudoDef.DATE_CREATED__FIELD_NAME, criteria.created_after)
            root = tb.and_terms(root, term)
        if criteria.created_before is not None:
            term = tb.lte_term(NotebookExperimentPseudoDef.DATE_CREATED__FIELD_NAME, criteria.created_before)
            root = tb.and_terms(root, term)

        if criteria.last_modified_after is not None:
            term = tb.gte_term(NotebookExperimentPseudoDef.LAST_MODIFIED_DATE__FIELD_NAME, criteria.last_modified_after)
            root = tb.and_terms(root, term)
        if criteria.last_modified_before is not None:
            term = tb.lte_term(NotebookExperimentPseudoDef.LAST_MODIFIED_DATE__FIELD_NAME, criteria.last_modified_before)
            root = tb.and_terms(root, term)

        if criteria.due_after is not None:
            term = tb.gte_term(NotebookExperimentPseudoDef.APPROVAL_DUE_DATE__FIELD_NAME, criteria.due_after)
            root = tb.and_terms(root, term)
        if criteria.due_before is not None:
            term = tb.lte_term(NotebookExperimentPseudoDef.APPROVAL_DUE_DATE__FIELD_NAME, criteria.due_before)
            root = tb.and_terms(root, term)

        report_builder.set_root_term(root)
        return report_builder.build_report_criteria()

    @staticmethod
    def get_experiments_for_report(context: UserIdentifier, report: CustomReportCriteria) -> list[ElnExperiment]:
        """
        Retrieve the ELN experiment objects for experiments whose notebook IDs appear in a custom report.

        :param context: The current webhook context or a user object to send requests from.
        :param report: A custom report that searches for ELN experiments, containing an "Experiment ID" column to
            query for the experiments of.
        :return: The ELN experiments that match the provided report.
        """
        user = AliasUtil.to_sapio_user(context)
        exp_ids: list[int] = []
        for row in CustomReportUtil.run_custom_report(user, report):
            exp_ids.append(row[NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME.field_name])
        if not exp_ids:
            return []

        criteria = ElnExperimentQueryCriteria(notebook_experiment_id_white_list=exp_ids)
        return ElnManager(user).get_eln_experiment_by_criteria(criteria)

    @staticmethod
    def all_notebook_experiments_term() -> RawReportTerm:
        """
        :return: A report term searching for all notebook experiments.
        """
        tb = TermBuilder(NotebookExperimentPseudoDef.DATA_TYPE_NAME)
        return tb.gte_term(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME, 0)

    @staticmethod
    def __get_record_experiment_relation_rows(user: SapioUser, data_type_name: str, record_ids: list[int] | None = None,
                                              exp_ids: list[int] | None = None) -> list[dict[str, FieldValue]]:
        """
        Return a list of dicts mapping \"RECORDID\" to the record id and \"EXPERIMENTID\" to the experiment id.
        At least one of record_ids and exp_ids should be provided.
        """
        assert (record_ids or exp_ids)

        report_builder = CustomReportBuilder(data_type_name)
        tb = report_builder.get_term_builder()
        if record_ids:
            records_term = tb.is_term("RecordId", record_ids)
        else:
            # Get all records of the given type
            records_term = tb.all_records_term()

        if exp_ids:
            exp_term = tb.is_term(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME, exp_ids,
                                  data_type=NotebookExperimentPseudoDef.DATA_TYPE_NAME)
        else:
            # Get all experiments
            exp_term = ExperimentReportUtil.all_notebook_experiments_term()

        root_term = tb.and_terms(records_term, exp_term)

        # Join records on the experiment entry records that correspond to them.
        records_entry_join = tb.compare_is_term(data_type_name,
                                                "RecordId",
                                                ExperimentEntryRecordPseudoDef.DATA_TYPE_NAME,
                                                ExperimentEntryRecordPseudoDef.RECORD_ID__FIELD_NAME)
        # Join entry records on the experiment entries they are in.
        experiment_entry_enb_entry_join = tb.compare_is_term(EnbEntryPseudoDef.DATA_TYPE_NAME,
                                                             EnbEntryPseudoDef.ENTRY_ID__FIELD_NAME,
                                                             ExperimentEntryRecordPseudoDef.DATA_TYPE_NAME,
                                                             ExperimentEntryRecordPseudoDef.ENTRY_ID__FIELD_NAME)
        # Join entries on the experiments they are in.
        enb_entry_experiment_join = tb.compare_is_term(NotebookExperimentPseudoDef.DATA_TYPE_NAME,
                                                       NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME,
                                                       EnbEntryPseudoDef.DATA_TYPE_NAME,
                                                       EnbEntryPseudoDef.EXPERIMENT_ID__FIELD_NAME)

        report_builder.set_root_term(root_term)
        report_builder.add_column("RecordId", FieldType.LONG)
        report_builder.add_column(NotebookExperimentPseudoDef.EXPERIMENT_ID__FIELD_NAME,
                                  data_type=NotebookExperimentPseudoDef.DATA_TYPE_NAME)
        report_builder.add_join(records_entry_join, ExperimentEntryRecordPseudoDef.DATA_TYPE_NAME)
        report_builder.add_join(experiment_entry_enb_entry_join, EnbEntryPseudoDef.DATA_TYPE_NAME)
        report_builder.add_join(enb_entry_experiment_join, NotebookExperimentPseudoDef.DATA_TYPE_NAME)
        return CustomReportUtil.run_custom_report(user, report_builder.build_report_criteria())



import unittest

from sapiopylib.rest.User import SapioUser

from sapiopycommons.multimodal.multimodal import MultiModalManager
from sapiopycommons.multimodal.multimodal_data import BioFileRegistrationRequest, BioFileType
from data_type_models import DNAPartModel
from data_type_models import ProteinPartModel

user = SapioUser(url="https://linux-vm:8443/webservice/api", verify_ssl_cert=False,
                 guid="66c2bea5-7cb2-4bfc-a413-304a3f4c3f33",
                 username="yqiao_api", password="Password1!")
reg_man = MultiModalManager(user)


class BioRegTest(unittest.TestCase):
    def test_genbank_registration(self):
        with open("test.gb", "r") as file_content:
            request = BioFileRegistrationRequest(DNAPartModel.DATA_TYPE_NAME, BioFileType.GENBANK, file_content.read())
            response = reg_man.register_bio(request)
            self.assertTrue(len(response.newRecordIdList) + len(response.oldRecordIdList) == 1)

    def test_fasta_registration(self):
        with open("kappa.chains.fasta", "r") as file_content:
            request = BioFileRegistrationRequest(DNAPartModel.DATA_TYPE_NAME, BioFileType.FASTA, file_content.read())
            response = reg_man.register_bio(request)
            self.assertTrue(len(response.newRecordIdList) + len(response.oldRecordIdList) == 15)

    def test_cif_registration(self):
        with open("AF-A0A009IHW8-F1-model_v4.cif", "r") as file_content:
            request = BioFileRegistrationRequest(ProteinPartModel.DATA_TYPE_NAME, BioFileType.CIF, file_content.read())
            response = reg_man.register_bio(request)
            self.assertTrue(len(response.newRecordIdList) + len(response.oldRecordIdList) == 1)



from __future__ import annotations

import json
import sys
import time
import traceback
from abc import abstractmethod
from logging import Logger

from sapiopylib.rest import UserManagerService, GroupManagerService, MessengerService
from sapiopylib.rest.AccessionService import AccessionManager
from sapiopylib.rest.CustomReportService import CustomReportManager
from sapiopylib.rest.DashboardManager import DashboardManager
from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.DataRecordManagerService import DataRecordManager
from sapiopylib.rest.DataService import DataManager
from sapiopylib.rest.DataTypeService import DataTypeManager
from sapiopylib.rest.ELNService import ElnManager
from sapiopylib.rest.PicklistService import PickListManager
from sapiopylib.rest.ReportManager import ReportManager
from sapiopylib.rest.SesssionManagerService import SessionManager
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.WebhookService import AbstractWebhookHandler
from sapiopylib.rest.pojo.Message import VeloxLogMessage, VeloxLogLevel
from sapiopylib.rest.pojo.webhook.ClientCallbackRequest import PopupType
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.pojo.webhook.WebhookEnums import WebhookEndpointType
from sapiopylib.rest.pojo.webhook.WebhookResult import SapioWebhookResult
from sapiopylib.rest.utils.DataTypeCacheManager import DataTypeCacheManager
from sapiopylib.rest.utils.FoundationAccessioning import FoundationAccessionManager
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager, RecordModelInstanceManager, \
    RecordModelRelationshipManager
from sapiopylib.rest.utils.recordmodel.ancestry import RecordModelAncestorManager
from sapiopylib.rest.utils.recordmodel.last_saved import LastSavedValueManager

from sapiopycommons.callbacks.callback_util import CallbackUtil
from sapiopycommons.eln.experiment_handler import ExperimentHandler
from sapiopycommons.general.directive_util import DirectiveUtil
from sapiopycommons.general.exceptions import SapioUserErrorException, SapioCriticalErrorException, \
    SapioUserCancelledException, SapioException, SapioDialogTimeoutException, MessageDisplayType
from sapiopycommons.general.sapio_links import SapioNavigationLinker
from sapiopycommons.general.time_util import TimeUtil
from sapiopycommons.recordmodel.record_handler import RecordHandler
from sapiopycommons.rules.eln_rule_handler import ElnRuleHandler
from sapiopycommons.rules.on_save_rule_handler import OnSaveRuleHandler
from sapiopycommons.webhook.webhook_context import ProcessQueueContext


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class CommonsWebhookHandler(AbstractWebhookHandler):
    """
    A subclass of AbstractWebhookHandler that provides additional quality of life features, including exception
    handling for special sapiopycommons exceptions, logging, easy access invocation type methods, and the context and
    record managers accessible through self.
    """
    logger: Logger

    _start_time: float
    _start_time_epoch: int

    user: SapioUser
    """The user who invoked this webhook. Used for authenticating requests back to the Sapio server."""
    group_name: str
    """The name of the group that the user invoked this webhook with."""
    user_utc_offset_seconds: int
    """The number of seconds that the user is offset from the UTC timezone. Able to be used with TimeUtil to display
    timestamps in the user's timezone."""
    context: SapioWebhookContext
    """The context from the server of this webhook's invocation."""

    # CR-47383: Include every manager from DataMgmtServer for easier access.
    acc_man: AccessionManager
    """A class for making requests to the accession webservice endpoints."""
    fnd_acc_man: FoundationAccessionManager
    """A class for making requests to the Foundations accession webservice endpoints."""
    custom_report_man: CustomReportManager
    """A class for making requests to the custom report webservice endpoints."""
    dash_man: DashboardManager
    """A class for making requests to the dashboard management webservice endpoints."""
    xml_data_man: DataManager
    """A class for making requests to the data record import/export via XML webservice endpoints."""
    dr_man: DataRecordManager
    """A class for making requests to the data record webservice endpoints."""
    dt_man: DataTypeManager
    """A class for making requests to the data type webservice endpoints."""
    eln_man: ElnManager
    """A class for making requests to the ELN management webservice endpoints."""
    group_man: GroupManagerService
    """A class for making requests to the group management webservice endpoints."""
    messenger: MessengerService
    """A class for making requests to the message webservice endpoints."""
    list_man: PickListManager
    """A class for making requests to the pick list webservice endpoints."""
    pdf_report_man: ReportManager
    """A class for making requests to the report webservice endpoints."""
    session_man: SessionManager
    """A class for making requests to the session management webservice endpoints."""
    user_man: UserManagerService
    """A class for making requests to the user management webservice endpoints."""

    rec_man: RecordModelManager
    """The record model manager. Used for committing record model changes to the system."""
    inst_man: RecordModelInstanceManager
    """The record model instance manager. Used for adding record models to the cache."""
    rel_man: RecordModelRelationshipManager
    """The record model relationship manager. Used for loading parent/child and side-link relationships between record
    models."""
    # FR-46329: Add the ancestor manager to CommonsWebhookHandler.
    an_man: RecordModelAncestorManager
    """The record model ancestor manager. Used for loading ancestor relationships between record models."""
    saved_vals_man: LastSavedValueManager
    """The last saved values manager. Used for determining what the record values were prior to this commit."""

    dt_cache: DataTypeCacheManager
    """A class that calls the same endpoints as the DataTypeManager (self.dt_man), except the results are cached so that
    repeated calls to the same function don't result in duplicate webservice calls. """
    rec_handler: RecordHandler
    """A class that behaves like a combination between the DataRecordManager and RecordModelInstanceManager, allowing
    you to query and wrap record as record models in a single function call, among other functions useful for dealing
    with record models."""
    callback: CallbackUtil
    """A class for making requests to the client callback webservice endpoints."""
    directive: DirectiveUtil
    """A class for making directives that redirect the user to a new webpage after this webhook returns a result."""
    exp_handler: ExperimentHandler | None
    """If this webhook was invoked from within an ELN experiment, this variable will be populated with an
    ExperimentHandler initialized from the context. """
    rule_handler: OnSaveRuleHandler | ElnRuleHandler | None
    """If this is an ELN or on save rule endpoint, this variable will be populated with an ElnRuleHandler or
    OnSaveRuleHandler, depending on the endpoint type."""
    custom_context: ProcessQueueContext | None
    """If this is a custom endpoint, this variable will be populated with an object that parses the custom context
    data."""

    # FR-47390: Allow for classes that extend CommonsWebhookHandler to change how exception messages are displayed
    # to the user be changing these variables instead of needing to override the exception handling functions.
    default_user_error_display_type: MessageDisplayType
    """The default message display type for user error exceptions. If a user error exception is thrown and doesn't
    specify a display type, this type will be used."""
    default_critical_error_display_type: MessageDisplayType
    """The default message display type for critical error exceptions. If a critical error exception is thrown and
    doesn't specify a display type, this type will be used."""
    default_dialog_timeout_display_type: MessageDisplayType
    """The default message display type for dialog timeout exceptions."""
    default_unexpected_error_display_type: MessageDisplayType
    """The default message display type for unexpected exceptions."""

    default_user_error_title: str
    """The default title to display to the user when a user error occurs. If a user error exception is thrown and
    doesn't specify a title, this title will be used."""
    default_critical_error_title: str
    """The default title to display to the user when a critical error occurs. If a critical error exception is thrown
    and doesn't specify a title, this title will be used."""
    default_dialog_timeout_title: str
    """The default title to display to the user when a dialog times out."""
    default_unexpected_error_title: str
    """The default title to display to the user when an unexpected exception occurs."""

    default_dialog_timeout_message: str
    """The default message to display to the user when a dialog times out."""
    default_unexpected_error_message: str
    """The default message to display to the user when an unexpected exception occurs."""

    def run(self, context: SapioWebhookContext) -> SapioWebhookResult:
        # Timestamps used for measuring performance.
        self._start_time = time.perf_counter()
        self._start_time_epoch = TimeUtil.now_in_millis()

        # Save the webhook context so that any function of this class can access it.
        self.context = context

        # Save the user and commonly sought after user information.
        self.user = context.user
        self.group_name = self.user.session_additional_data.current_group_name
        self.user_utc_offset_seconds = self.user.session_additional_data.utc_offset_seconds

        # Get the logger from the user.
        self.logger = self.user.logger

        # Initialize basic manager classes from sapiopylib.
        self.acc_man = DataMgmtServer.get_accession_manager(self.user)
        self.custom_report_man = DataMgmtServer.get_custom_report_manager(self.user)
        self.dash_man = DataMgmtServer.get_dashboard_manager(self.user)
        self.xml_data_man = DataMgmtServer.get_data_manager(self.user)
        self.dr_man = context.data_record_manager
        self.dt_man = DataMgmtServer.get_data_type_manager(self.user)
        self.eln_man = context.eln_manager
        self.group_man = DataMgmtServer.get_group_manager(self.user)
        self.messenger = DataMgmtServer.get_messenger(self.user)
        self.list_man = DataMgmtServer.get_picklist_manager(self.user)
        self.pdf_report_man = DataMgmtServer.get_report_manager(self.user)
        self.session_man = DataMgmtServer.get_session_manager(self.user)
        self.user_man = DataMgmtServer.get_user_manager(self.user)

        # Initialize record model managers.
        self.rec_man = RecordModelManager(self.user)
        self.inst_man = self.rec_man.instance_manager
        self.rel_man = self.rec_man.relationship_manager
        self.an_man = self.rec_man.ancestor_manager
        self.saved_vals_man = self.rec_man.last_saved_manager

        # Initialize more complex classes from sapiopylib and sapiopycommons.
        self.fnd_acc_man = FoundationAccessionManager(self.user)
        self.dt_cache = DataTypeCacheManager(self.user)
        self.rec_handler = RecordHandler(context)
        self.callback = CallbackUtil(context)
        self.directive = DirectiveUtil(context)
        if context.eln_experiment is not None:
            self.exp_handler = ExperimentHandler(context)
        else:
            self.exp_handler = None
        if self.is_on_save_rule():
            self.rule_handler = OnSaveRuleHandler(context)
        elif self.is_eln_rule():
            self.rule_handler = ElnRuleHandler(context)
        else:
            self.rule_handler = None
        if self.is_custom():
            self.custom_context = ProcessQueueContext(context)
        else:
            self.custom_context = None

        # Set the default display types, titles, and messages for each type of exception that can display a message.
        self.default_user_error_display_type = MessageDisplayType.TOASTER_WARNING
        self.default_critical_error_display_type = MessageDisplayType.DISPLAY_ERROR
        self.default_dialog_timeout_display_type = MessageDisplayType.OK_DIALOG
        self.default_unexpected_error_display_type = MessageDisplayType.TOASTER_WARNING

        self.default_user_error_title = ""
        self.default_critical_error_title = ""
        self.default_dialog_timeout_title = "Dialog Timeout"
        self.default_unexpected_error_title = ""

        self.default_dialog_timeout_message = ("You have remained idle for too long and this dialog has timed out. "
                                               "Close and re-initiate it to continue.")
        self.default_unexpected_error_message = ("Unexpected error occurred during webhook execution. Please contact "
                                                 "Sapio support.")

        # Wrap the execution of each webhook in a try/catch. If an exception occurs, handle any special sapiopycommons
        # exceptions. Otherwise, return a generic message stating that an error occurred.
        try:
            self.initialize(context)
            result = self.execute(context)
            if result is None:
                raise SapioException("Your execute function returned a None result! Don't forget your return statement!")
            return result
        except SapioUserErrorException as e:
            return self.handle_user_error_exception(e)
        except SapioCriticalErrorException as e:
            return self.handle_critical_error_exception(e)
        except SapioUserCancelledException as e:
            return self.handle_user_cancelled_exception(e)
        except SapioDialogTimeoutException as e:
            return self.handle_dialog_timeout_exception(e)
        except Exception as e:
            return self.handle_unexpected_exception(e)

    def initialize(self, context: SapioWebhookContext) -> None:
        """
        A function that can be optionally overridden by your webhooks to initialize additional instance variables,
        or set up whatever else you wish to set up before the execute function is ran. Default behavior does nothing.
        """
        pass

    @abstractmethod
    def execute(self, context: SapioWebhookContext) -> SapioWebhookResult:
        """
        The business logic of the webhook, implemented in all subclasses that are called as endpoints.
        """
        pass

    # CR-46153: Make CommonsWebhookHandler exception handling more easily overridable by splitting them out of
    # the run method and into their own functions.
    def handle_user_error_exception(self, e: SapioUserErrorException) -> SapioWebhookResult:
        """
        Handle a SapioUserErrorException.

        Default behavior returns a false result and the error message as display text in a webhook result.

        :param e: The exception that was raised.
        :return: A SapioWebhookResult to end the webhook session with.
        """
        result: SapioWebhookResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        self.log_error(traceback.format_exc())
        self.handle_user_error_exception_extra(e)

        display_type: MessageDisplayType = e.display_type if e.display_type else self.default_user_error_display_type
        title: str = e.title if e.title is not None else self.default_user_error_title
        return self._display_exception(e.msg, display_type, title)

    def handle_user_error_exception_extra(self, e: SapioUserErrorException) -> None:
        """
        An additional function that can be overridden to provide extra behavior when a SapioUserErrorException is thrown.
        Default behavior does nothing.
        """
        pass

    def handle_critical_error_exception(self, e: SapioCriticalErrorException) -> SapioWebhookResult:
        """
        Handle a SapioCriticalErrorException.

        Default behavior makes a display_error client callback with the error message and returns a false result.

        :param e: The exception that was raised.
        :return: A SapioWebhookResult to end the webhook session with.
        """
        result: SapioWebhookResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        self.log_error(traceback.format_exc())
        self.handle_critical_error_exception_extra(e)

        display_type: MessageDisplayType = e.display_type if e.display_type else self.default_critical_error_display_type
        title: str = e.title if e.title is not None else self.default_critical_error_title
        return self._display_exception(e.msg, display_type, title)

    def handle_critical_error_exception_extra(self, e: SapioCriticalErrorException) -> None:
        """
        An additional function that can be overridden to provide extra behavior when a SapioCriticalErrorException is
        thrown. Default behavior does nothing.
        """
        pass

    def handle_user_cancelled_exception(self, e: SapioUserCancelledException) -> SapioWebhookResult:
        """
        Handle a SapioUserCancelledException.

        Default behavior simply ends the webhook session with a true result (since the user cancelling is a valid
        action).

        :param e: The exception that was raised.
        :return: A SapioWebhookResult to end the webhook session with.
        """
        result: SapioWebhookResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        self.handle_user_cancelled_exception_extra(e)
        # FR-47390: Return a False result for user cancelled exceptions so that transactional webhooks cancel the
        # commit.
        return SapioWebhookResult(False)

    def handle_user_cancelled_exception_extra(self, e: SapioUserCancelledException) -> None:
        """
        An additional function that can be overridden to provide extra behavior when a SapioUserCancelledException is
        thrown. Default behavior does nothing.
        """
        pass

    def handle_dialog_timeout_exception(self, e: SapioDialogTimeoutException) -> SapioWebhookResult:
        """
        Handle a SapioDialogTimeoutException.

        Default behavior displays an OK popup notifying the user that the dialog has timed out and returns a false
        webhook result.

        :param e: The exception that was raised.
        :return: A SapioWebhookResult to end the webhook session with.
        """
        result: SapioWebhookResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        self.handle_dialog_timeout_exception_extra(e)
        return self._display_exception(self.default_dialog_timeout_message,
                                       self.default_dialog_timeout_display_type,
                                       self.default_dialog_timeout_title)

    def handle_dialog_timeout_exception_extra(self, e: SapioDialogTimeoutException) -> None:
        """
        An additional function that can be overridden to provide extra behavior when a SapioDialogTimeoutException is
        thrown. Default behavior does nothing.
        """
        pass

    def handle_unexpected_exception(self, e: Exception) -> SapioWebhookResult:
        """
        Handle a generic exception which isn't one of the handled Sapio exceptions.

        Default behavior returns a false webhook result with a generic error message as display text informing the user
        to contact Sapio support. Additionally, the stack trace of the exception that was thrown is logged to the
        execution log for the webhook call in the system.

        :param e: The exception that was raised.
        :return: A SapioWebhookResult to end the webhook session with.
        """
        result: SapioWebhookResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        msg: str = traceback.format_exc()
        self.log_error(msg, True)
        # FR-47079: Also log all unexpected exception messages to the webhook execution log within the platform.
        self.log_error_to_webhook_execution_log(msg, True)
        self.handle_unexpected_exception_extra(e)
        return self._display_exception(self.default_unexpected_error_message,
                                       self.default_unexpected_error_display_type,
                                       self.default_unexpected_error_title)

    def handle_unexpected_exception_extra(self, e: Exception) -> None:
        """
        An additional function that can be overridden to provide extra behavior when a generic exception is thrown.
        Default behavior does nothing.
        """
        pass

    def handle_any_exception(self, e: Exception) -> SapioWebhookResult | None:
        """
        An exception handler which runs regardless of the type of exception that was raised. Can be used to "rollback"
        the client if an error occurs. Default behavior does nothing and returns None.

        :param e: The exception that was raised.
        :return: An optional SapioWebhookResult. May return a custom message to the client that wouldn't have been
            sent by one of the normal exception handlers, or may return None if no result needs returned. If a result is
            returned, then the default behavior of other exception handlers is skipped.
        """
        pass

    def display_message(self, msg: str, display_type: MessageDisplayType, title: str = "") -> bool:
        """
        Display a message to the user. The form that the message takes depends on the display type.

        :param msg: The message to display to the user.
        :param display_type: The manner in which the message should be displayed.
        :param title: If the display type is able to have a title, this is the title that will be displayed.
        :return: True if the message was displayed. False if the message could not be displayed (because this
            webhook can't send client callbacks).
        """
        if not self.can_send_client_callback():
            return False
        if display_type == MessageDisplayType.TOASTER_SUCCESS:
            self.callback.toaster_popup(msg, title, PopupType.Success)
        elif display_type == MessageDisplayType.TOASTER_INFO:
            self.callback.toaster_popup(msg, title, PopupType.Info)
        elif display_type == MessageDisplayType.TOASTER_WARNING:
            self.callback.toaster_popup(msg, title, PopupType.Warning)
        elif display_type == MessageDisplayType.TOASTER_ERROR:
            self.callback.toaster_popup(msg, title, PopupType.Error)
        elif display_type == MessageDisplayType.OK_DIALOG:
            self.callback.ok_dialog(title, msg)
        elif display_type == MessageDisplayType.DISPLAY_INFO:
            self.callback.display_info(msg)
        elif display_type == MessageDisplayType.DISPLAY_WARNING:
            self.callback.display_warning(msg)
        elif display_type == MessageDisplayType.DISPLAY_ERROR:
            self.callback.display_error(msg)
        return True

    def _display_exception(self, msg: str, display_type: MessageDisplayType, title: str) -> SapioWebhookResult:
        """
        Display an exception message to the user and return a webhook result to end the webhook invocation.
        This handles the cases where the webhook invocation type is incapable of sending client callbacks and must
        instead return the message in the webhook result, and the case where the display type is an OK dialog, which
        may potentially cause a dialog timeout exception.
        """
        # If the display type is an OK dialog, then we need to handle the dialog timeout exception that could be thrown.
        try:
            # Set the dialog timeout to something low as to not hog the connection.
            self.callback.set_dialog_timeout(60)
            # If this invocation type can't send client callbacks, fallback to sending the message in the result.
            if self.display_message(msg, display_type, title):
                return SapioWebhookResult(False)
            return SapioWebhookResult(False, display_text=msg)
        except SapioDialogTimeoutException:
            return SapioWebhookResult(False)

    def log_info(self, msg: str) -> None:
        """
        Write an info message to the webhook server log. Log destination is stdout. This message will include
        information about the user's group, their location in the system, the webhook invocation type, and other
        important information that can be gathered from the context that is useful for debugging.
        """
        self.logger.info(self._format_log(msg, "log_info call"))

    def log_error(self, msg: str, is_exception: bool = False) -> None:
        """
        Write an info message to the webhook server log. Log destination is stdout. This message will include
        information about the user's group, their location in the system, the webhook invocation type, and other
        important information that can be gathered from the context that is useful for debugging.
        """
        # PR-46209: Use logger.error instead of logger.info when logging errors.
        self.logger.error(self._format_log(msg, "log_error call", is_exception))

    def log_error_to_webhook_execution_log(self, msg: str, is_exception: bool = False) -> None:
        """
        Write an error message to the platform's webhook execution log. This can be reviewed by navigating to the
        webhook configuration where the webhook that called this function is defined and clicking the "View Log"
        button. From there, select one of the rows for the webhook executions and click "Download Log" from the right
        side table.
        """
        msg = self._format_log(msg, "Error occurred during webhook execution.", is_exception)
        self.messenger.log_message(VeloxLogMessage(msg, VeloxLogLevel.ERROR, self.__class__.__name__))

    def _format_log(self, msg: str, prefix: str | None = None, is_exception: bool = False) -> str:
        """
        Given a message to log, populate it with some metadata about this particular webhook execution, including
        the group of the user and the invocation type of the webhook call.
        """
        # Start the message with the provided prefix.
        message: str = prefix + "\n" if prefix else ""

        # Construct a summary of the current state of this webhook.
        message += f"{WebhookStateSummary(self, is_exception)}\n"

        # End the message with the provided msg parameter.
        message += msg
        return message

    @property
    def start_time(self) -> float:
        """
        :return: The time that this webhook was invoked, represented in seconds. This time comes from a performance
            counter and is not guaranteed to correspond to a date. Only use in comparison to other performance counters.
        """
        return self._start_time

    @property
    def start_time_millis(self) -> int:
        """
        :return: The epoch timestamp in milliseconds for the time that this webhook was invoked.
        """
        return self._start_time_epoch

    def elapsed_time(self) -> float:
        """
        :return: The number of seconds that have elapsed since this webhook was invoked to the time that this function
            is called. Measures using a performance counter to a high degree of accuracy.
        """
        return time.perf_counter() - self._start_time

    def elapsed_time_millis(self) -> int:
        """
        :return: The number of milliseconds that have elapsed since this webhook was invoked to the time that this
            function is called.
        """
        return TimeUtil.now_in_millis() - self._start_time_epoch

    def is_main_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a main toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.ACTIONMENU

    def is_form_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a data record form toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.FORMTOOLBAR

    def is_table_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a data record table toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.TABLETOOLBAR

    def is_temp_form_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a temporary data record form toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.TEMP_DATA_FORM_TOOLBAR

    def is_temp_table_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a temporary data record table toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.TEMP_DATA_TABLE_TOOLBAR

    def is_eln_rule(self) -> bool:
        """
        :return: True if this endpoint was invoked as an ELN rule action.
        """
        return self.context.end_point_type == WebhookEndpointType.VELOXELNRULEACTION

    def is_on_save_rule(self) -> bool:
        """
        :return: True if this endpoint was invoked as an on save rule action.
        """
        return self.context.end_point_type == WebhookEndpointType.VELOX_RULE_ACTION
        # TODO: This VELOXONSAVERULEACTION endpoint type exists, but I don't see it actually getting sent by on save
        #  rule action invocations, instead seeing the above VELOX_RULE_ACTION type. Probably worth investigation.
        # return self.context.end_point_type == WebhookEndpointType.VELOXONSAVERULEACTION

    def is_eln_main_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as an ELN main toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.NOTEBOOKEXPERIMENTMAINTOOLBAR

    def is_eln_entry_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as an ELN entry toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.EXPERIMENTENTRYTOOLBAR

    def is_selection_list(self) -> bool:
        """
        :return: True if this endpoint was invoked as a selection list populator.
        """
        return self.context.end_point_type == WebhookEndpointType.SELECTIONDATAFIELD

    def is_report_builder(self) -> bool:
        """
        :return: True if this endpoint was invoked as a report builder template data populator.
        """
        return self.context.end_point_type == WebhookEndpointType.REPORT_BUILDER_TEMPLATE_DATA_POPULATOR

    def is_scheduled_action(self) -> bool:
        """
        :return: True if this endpoint was invoked as a scheduled action.
        """
        return self.context.end_point_type == WebhookEndpointType.SCHEDULEDPLUGIN

    def is_action_button_field(self) -> bool:
        """
        :return: True if this endpoint was invoked as an action button field.
        """
        return self.context.end_point_type == WebhookEndpointType.ACTIONDATAFIELD

    def is_action_text_field(self) -> bool:
        """
        :return: True if this endpoint was invoked as an action text field.
        """
        return self.context.end_point_type == WebhookEndpointType.ACTION_TEXT_FIELD

    def is_custom(self) -> bool:
        """
        :return: True if this endpoint was invoked from a custom point, such as a custom queue.
        """
        return self.context.end_point_type == WebhookEndpointType.CUSTOM

    def is_calendar_event_click_handler(self) -> bool:
        """
        :return: True if this endpoint was invoked from a calendar event click handler.
        """
        return self.context.end_point_type == WebhookEndpointType.CALENDAR_EVENT_CLICK_HANDLER

    def is_eln_menu_grabber(self) -> bool:
        """
        :return: True if this endpoint was invoked as a notebook entry grabber.
        """
        return self.context.end_point_type == WebhookEndpointType.NOTEBOOKEXPERIMENTGRABBER

    def is_conversation_bot(self) -> bool:
        """
        :return: True if this endpoint was invoked as from a conversation bot.
        """
        return self.context.end_point_type == WebhookEndpointType.CONVERSATION_BOT

    def is_multi_data_type_table_toolbar(self) -> bool:
        """
        :return: True if this endpoint was invoked as a multi data type table toolbar button.
        """
        return self.context.end_point_type == WebhookEndpointType.REPORTTOOLBAR

    def can_send_client_callback(self) -> bool:
        """
        :return: Whether client callbacks and directives can be sent from this webhook's endpoint type.
        """
        return self.context.is_client_callback_available


# FR-47390: Move the gathering of webhook information out of log_error_to_webhook_execution_log and into its own class.
class WebhookStateSummary:
    """
    A class that summarizes the state of a webhook at the time that it is created. This class is useful for logging
    information about the webhook invocation to the execution log.
    """
    username: str
    """The username of the user who invoked the webhook."""
    user_group: str
    """The group that the user is currently in."""
    start_timestamp: int
    """The epoch timestamp in milliseconds for when the webhook was invoked."""
    start_utc_time: str
    """The time that the webhook was invoked in UTC."""
    start_server_time: str | None
    """The time that the webhook was invoked on the server, if the TimeUtil class has a default timezone set."""
    start_user_time: str
    """The time that the webhook was invoked for the user, adjusted for their timezone."""
    timestamp: int
    """The current epoch timestamp in milliseconds."""
    utc_time: str
    """The current time in UTC."""
    server_time: str | None
    """The current time on the webhook server, if the TimeUtil class has a default timezone set."""
    user_time: str
    """The current time for the user, adjusted for their timezone."""
    invocation_type: str
    """The type of endpoint that this webhook was invoked from."""
    class_name: str
    """The name of the class that this webhook is an instance of."""
    link: str | None
    """A link to the location that the webhook was invoked from, if applicable."""
    exc_summary: str | None
    """A summary of the exception that occurred, if this summary is being created for an exception."""

    def __init__(self, webhook: CommonsWebhookHandler, summarize_exception: bool = False):
        """
        :param webhook: The webhook that this summary is being created for.
        :param summarize_exception: If true, then this summary will include information about the most recent exception
            that occurred during the execution of the webhook.
        """
        # User information.
        self.username = webhook.user.username
        self.user_group = webhook.group_name

        # Time information.
        fmt: str = "%Y-%m-%d %H:%M:%S.%f"
        self.start_timestamp = webhook.start_time_millis
        self.start_utc_time = TimeUtil.millis_to_format(self.start_timestamp, fmt, "UTC")
        self.start_server_time = None
        self.start_user_time = TimeUtil.millis_to_format(self.start_timestamp, fmt, webhook.user_utc_offset_seconds)

        self.timestamp = TimeUtil.now_in_millis()
        self.utc_time = TimeUtil.millis_to_format(self.timestamp, fmt, "UTC")
        self.server_time = None
        self.user_time = TimeUtil.millis_to_format(self.timestamp, fmt, webhook.user_utc_offset_seconds)

        if TimeUtil.get_default_timezone():
            self.start_server_time = TimeUtil.millis_to_format(self.start_timestamp, fmt)
            self.server_time = TimeUtil.millis_to_format(self.timestamp, fmt)

        # Webhook invocation information.
        self.invocation_type = webhook.context.end_point_type.display_name
        self.class_name = webhook.__class__.__name__

        # User location information.
        self.link = None
        context = webhook.context
        navigator = SapioNavigationLinker(context)
        if context.eln_experiment is not None:
            self.link = navigator.experiment(context.eln_experiment)
        elif context.base_data_record:
            self.link = navigator.data_record(context.base_data_record)
        elif context.data_record and not context.data_record_list:
            self.link = navigator.data_record(context.data_record)

        # If this is logging an exception, get the system's info on the most recent exception and produce a summary.
        self.exc_summary = None
        if summarize_exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()

            # If this is a SapioServerException, then it can contain json with the exception message from the server.
            # This provides a more useful summary than args[0] does, so use it instead.
            exception_name: str = exc_type.__name__
            if exception_name == "SapioServerException":
                # Sometimes the SapioServerException is HTML instead of JSON. If that's the case, just try/catch the
                # failure to parse the JSON and continue to use args[0] as the exception message.
                try:
                    exc_str: str = str(exc_value)
                    exception_msg = json.loads(exc_str[exc_str.find("{"):]).get("message")
                except Exception:
                    exception_msg = exc_value.args[0]
            else:
                # For all other exceptions, assume that the first argument is a message.
                exception_msg = exc_value.args[0]

            self.exc_summary = f"{exception_name}: {exception_msg}"
            del (exc_type, exc_value, exc_traceback)

    def __str__(self) -> str:
        message: str = ""

        # Record the time that the webhook was started and this state was created.
        message += "Webhook Invocation Time:\n"
        message += f"\tUTC: {self.start_utc_time}\n"
        if TimeUtil.get_default_timezone() is not None:
            message += f"\tServer: {self.start_server_time}\n"
        message += f"\tUser: {self.start_user_time}\n"

        message += "Current Time:\n"
        message += f"\tUTC: {self.utc_time}\n"
        if TimeUtil.get_default_timezone() is not None:
            message += f"\tServer: {self.server_time}\n"
        message += f"\tUser: {self.user_time}\n"

        # Record information about the user and how the webhook was invoked.
        message += f"Username: {self.username}\n"
        message += f"User group: {self.user_group}\n"
        message += f"Webhook invocation type: {self.invocation_type}\n"
        message += f"Class name: {self.class_name}\n"

        # If we're able to, provide a link to the location that the error occurred at.
        if self.link:
            message += f"User location: {self.link}\n"

        # If this state summary is for an exception, include the exception summary.
        if self.exc_summary:
            message += f"{self.exc_summary}\n"

        return message



from typing import Iterable, cast

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import CustomReportCriteria, CustomReport
from sapiopylib.rest.pojo.webhook.WebhookDirective import HomePageDirective, FormDirective, TableDirective, \
    CustomReportDirective, ElnExperimentDirective, ExperimentEntryDirective

from sapiopycommons.general.aliases import SapioRecord, AliasUtil, ExperimentIdentifier, ExperimentEntryIdentifier, \
    UserIdentifier
from sapiopycommons.general.custom_report_util import CustomReportUtil


# FR-47392: Create a DirectiveUtil class to simplify the creation of directives.
class DirectiveUtil:
    """
    DirectiveUtil is a class for creating webhook directives. The utility functions reduce the provided variables
    down to the exact type that the directives require, removing the need for the caller to handle the conversion.
    """
    user: SapioUser

    def __init__(self, context: UserIdentifier):
        """
        :param context: The current webhook context or a user object to send requests from.
        """
        self.user = AliasUtil.to_sapio_user(context)

    @staticmethod
    def homepage() -> HomePageDirective:
        """
        :return: A directive that sends the user back to their home page.
        """
        return HomePageDirective()

    @staticmethod
    def record_form(record: SapioRecord) -> FormDirective:
        """
        :param record: A record in the system.
        :return: A directive that sends the user to a specific data record form.
        """
        return FormDirective(AliasUtil.to_data_record(record))

    @staticmethod
    def record_table(records: Iterable[SapioRecord]) -> TableDirective:
        """
        :param records: A list of records in the system.
        :return: A directive that sends the user to a table of data records.
        """
        return TableDirective(AliasUtil.to_data_records(records))

    @staticmethod
    def record_adaptive(records: Iterable[SapioRecord]) -> TableDirective | FormDirective:
        """
        :param records: A list of records in the system.
        :return: A directive that sends the user to a table of data records if there are multiple records,
            or a directive that sends the user to a specific data record form if there is only one record.
        """
        records: list[SapioRecord] = list(records)
        if len(records) == 1:
            return DirectiveUtil.record_form(records[0])
        return DirectiveUtil.record_table(records)

    def custom_report(self, report: CustomReport | CustomReportCriteria | str) -> CustomReportDirective:
        """
        :param report: A custom report, the criteria for a custom report, or the name of a system report.
        :return: A directive that sends the user to the results of the provided custom report.
        """
        if isinstance(report, str):
            report: CustomReport = CustomReportUtil.get_system_report_criteria(self.user, report)
        return CustomReportDirective(cast(CustomReport, report))

    @staticmethod
    def eln_experiment(experiment: ExperimentIdentifier) -> ElnExperimentDirective:
        """
        :param experiment: An identifier for an experiment.
        :return: A directive that sends the user to the ELN experiment.
        """
        return ElnExperimentDirective(AliasUtil.to_notebook_id(experiment))

    @staticmethod
    def eln_entry(experiment: ExperimentIdentifier, entry: ExperimentEntryIdentifier) -> ExperimentEntryDirective:
        """
        :param experiment: An identifier for an experiment.
        :param entry: An identifier for an entry in the experiment.
        :return: A directive that sends the user to the provided experiment entry within its ELN experiment.
        """
        return ExperimentEntryDirective(AliasUtil.to_notebook_id(experiment), AliasUtil.to_entry_id(entry))



from __future__ import annotations

from weakref import WeakValueDictionary

from sapiopylib.rest.User import SapioUser
from databind.json import dumps

from sapiopycommons.flowcyto.flowcyto_data import FlowJoWorkspaceInputJson, UploadFCSInputJson, \
    ComputeFlowStatisticsInputJson


class FlowCytoManager:
    """
    This manager includes flow cytometry analysis tools that would require FlowCyto license to use.
    """
    _user: SapioUser

    __instances: WeakValueDictionary[SapioUser, FlowCytoManager] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, user: SapioUser):
        """
        Observes singleton pattern per record model manager object.

        :param user: The user that will make the webservice request to the application.
        """
        obj = cls.__instances.get(user)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[user] = obj
        return obj

    def __init__(self, user: SapioUser):
        if self.__initialized:
            return
        self._user = user
        self.__initialized = True

    def create_flowjo_workspace(self, workspace_input: FlowJoWorkspaceInputJson) -> int:
        """
        Create FlowJo Workspace and return the workspace record ID of workspace root record,
        after successful creation.
        :param workspace_input: the request data payload.
        :return: The new workspace record ID.
        """
        payload = dumps(workspace_input, FlowJoWorkspaceInputJson)
        response = self._user.plugin_post("flowcyto/workspace", payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return int(response.json())

    def upload_fcs_for_sample(self, upload_input: UploadFCSInputJson) -> int:
        """
        Upload FCS file as root of the sample FCS.
        :param upload_input: The request data payload
        :return: The root FCS file uploaded under sample.
        """
        payload = dumps(upload_input, UploadFCSInputJson)
        response = self._user.plugin_post("flowcyto/fcs", payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return int(response.json())

    def compute_statistics(self, stat_compute_input: ComputeFlowStatisticsInputJson) -> list[int]:
        """
        Requests to compute flow cytometry statistics.
        The children are of type FCSStatistic.
        If the FCS files have not been evaluated yet,
        then the lazy evaluation will be performed immediately prior to computing statistics, which can take longer.
        If any new statistics are computed as children of FCS, they will be returned in the result record id list.
        Note: if input has multiple FCS files, the client should try to get parent FCS file from each record to figure out which one is for which FCS.
        :param stat_compute_input:
        :return:
        """
        payload = dumps(stat_compute_input, ComputeFlowStatisticsInputJson)
        response = self._user.plugin_post("flowcyto/statistics", payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return list(response.json())



from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrapperField


class SystemFields:
    RECORD_ID__FIELD = WrapperField("RecordId", FieldType.LONG)
    DATA_RECORD_NAME__FIELD = WrapperField("DataRecordName", FieldType.IDENTIFIER)
    CREATED_BY__FIELD = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD = WrapperField("DateCreated", FieldType.DATE)
    LAST_MODIFIED_BY__FIELD = WrapperField("VeloxLastModifiedBy", FieldType.STRING)
    LAST_MODIFIED_DATE__FIELD = WrapperField("VeloxLastModifiedDate", FieldType.DATE)


class ProcessQueueItemFields:
    DATA_TYPE_NAME: str = 'ProcessQueueItem'
    ASSIGNED_TO__FIELD = WrapperField('AssignedTo', FieldType.SELECTION)
    DATA_RECORD_ID__FIELD = WrapperField('DataRecordId', FieldType.LONG)
    DATA_TYPE_NAME__FIELD = WrapperField('DataTypeName', FieldType.STRING)
    DURATION_IN_QUEUE__FIELD = WrapperField('DurationInQueue', FieldType.LONG)
    EXPERIMENT__FIELD = WrapperField('Experiment', FieldType.SIDE_LINK)
    LAUNCHED_DATE__FIELD = WrapperField('LaunchedDate', FieldType.DATE)
    PROCESS_HEADER_NAME__FIELD = WrapperField('ProcessHeaderName', FieldType.STRING)
    SCHEDULED_DATE__FIELD = WrapperField('ScheduledDate', FieldType.DATE)
    SHOW_IN_QUEUE__FIELD = WrapperField('ShowInQueue', FieldType.BOOLEAN)
    WORKFLOW_HEADER_NAME__FIELD = WrapperField('WorkflowHeaderName', FieldType.STRING)


class ProcessWorkflowTrackingFields:
    DATA_TYPE_NAME: str = 'ProcessWorkflowTracking'
    ACTIVE_WORKFLOW_ID__FIELD = WrapperField("ActiveWorkflowId", FieldType.LONG)
    BRANCH_LONG_ID__FIELD = WrapperField("BranchLongId", FieldType.LONG)
    END_DATE__FIELD = WrapperField("EndDate", FieldType.DATE)
    END_HOUR__FIELD = WrapperField("EndHour", FieldType.LONG)
    END_MINUTE_TIME__FIELD = WrapperField("EndMinuteTime", FieldType.LONG)
    EXPECTED_QUEUE_TIME__FIELD = WrapperField("ExpectedQueueTime", FieldType.DOUBLE)
    EXPECTED_TAT__FIELD = WrapperField("ExpectedTAT", FieldType.DOUBLE)
    PROCESS_QUEUE_ITEM__FIELD = WrapperField("ProcessQueueItem", FieldType.SIDE_LINK)
    PROCESS_STEP_NUMBER__FIELD = WrapperField("ProcessStepNumber", FieldType.LONG)
    QUEUED_HOURS__FIELD = WrapperField("QueuedHours", FieldType.LONG)
    QUEUED_MINUTES__FIELD = WrapperField("QueuedMinutes", FieldType.LONG)
    QUEUED_TIME__FIELD = WrapperField("QueuedTime", FieldType.DOUBLE)
    QUEUE_START_DATE__FIELD = WrapperField("QueueStartDate", FieldType.DATE)
    QUEUE_START_HOUR__FIELD = WrapperField("QueueStartHour", FieldType.LONG)
    QUEUE_START_MINUTE_TIME__FIELD = WrapperField("QueueStartMinuteTime", FieldType.LONG)
    SAMPLE_RECORD_ID__FIELD = WrapperField("SampleRecordId", FieldType.LONG)
    START_DATE__FIELD = WrapperField("StartDate", FieldType.DATE)
    START_HOUR__FIELD = WrapperField("StartHour", FieldType.LONG)
    START_MINUTE_TIME__FIELD = WrapperField("StartMinuteTime", FieldType.LONG)
    STATUS__FIELD = WrapperField("Status", FieldType.PICKLIST)
    TURN_AROUND_HOURS__FIELD = WrapperField("TurnAroundHours", FieldType.LONG)
    TURN_AROUND_MINUTES__FIELD = WrapperField("TurnAroundMinutes", FieldType.LONG)
    WORKFLOW_END_USER_ID__FIELD = WrapperField("WorkflowEndUserId", FieldType.STRING)
    WORKFLOW_EXPECTED_QUEUE_TIME__FIELD = WrapperField("WorkflowExpectedQueueTime", FieldType.STRING)
    WORKFLOW_EXPECTED_TAT__FIELD = WrapperField("WorkflowExpectedTAT", FieldType.STRING)
    WORKFLOW_EXPECTED_TOTAL_TAT__FIELD = WrapperField("WorkflowExpectedTotalTAT", FieldType.DOUBLE)
    WORKFLOW_ID_NUMBER__FIELD = WrapperField("WorkflowIdNumber", FieldType.LONG)
    WORKFLOW_NAME__FIELD = WrapperField("WorkflowName", FieldType.STRING)
    WORKFLOW_PROCESS_TAT__FIELD = WrapperField("WorkflowProcessTAT", FieldType.DOUBLE)
    WORKFLOW_START_USER_ID__FIELD = WrapperField("WorkflowStartUserId", FieldType.STRING)
    WORKFLOW_TAT__FIELD = WrapperField("WorkflowTAT", FieldType.DOUBLE)
    WORKFLOW_VERSION__FIELD = WrapperField("WorkflowVersion", FieldType.LONG)


import os
import unittest

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.DataRecord import DataRecord
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager

from sapiopycommons.flowcyto.flow_cyto import FlowCytoManager
from data_type_models import *

from sapiopycommons.flowcyto.flowcyto_data import UploadFCSInputJson, FlowJoWorkspaceInputJson, \
    ComputeFlowStatisticsInputJson, ChannelStatisticsParameterJSON, ChannelStatisticType

user = SapioUser(url="https://linux-vm:8443/webservice/api", verify_ssl_cert=False,
                 guid="66c2bea5-7cb2-4bfc-a413-304a3f4c3f33",
                 username="yqiao_api", password="Password1!")
flow_man = FlowCytoManager(user)
data_record_manager = DataMgmtServer.get_data_record_manager(user)
rec_man = RecordModelManager(user)
inst_man = rec_man.instance_manager
data_dir = "flowcyto"

def _create_attachment_record_of_content(file_path: str) -> DataRecord:
    record = data_record_manager.add_data_record(AttachmentModel.DATA_TYPE_NAME)
    file_name: str = os.path.basename(file_path)
    with open(file_path, "rb") as file_content:
        data_record_manager.set_attachment_data(record, file_name, file_content)
    return record

# NOTE: This unit test will run for a few minutes...

class FlowCytoTest(unittest.TestCase):

    def test_sample_import_and_stats(self):
        file_name = "COVID19_W_001_O.fcs"
        file_path = data_dir + "/" + file_name
        sample_rec: DataRecord = data_record_manager.add_data_record(SampleModel.DATA_TYPE_NAME)
        fcs_attachment_rec = _create_attachment_record_of_content(file_path)
        response_record_id = flow_man.upload_fcs_for_sample(
            UploadFCSInputJson(sample_rec.data_type_name, sample_rec.record_id, file_name,
                               fcs_attachment_rec.data_type_name, fcs_attachment_rec.record_id))
        self.assertTrue(response_record_id == sample_rec.record_id)
        children_fcs_recs = data_record_manager.get_children(sample_rec.record_id, FCSFileModel.DATA_TYPE_NAME)
        input_json = ComputeFlowStatisticsInputJson([x.record_id for x in children_fcs_recs], [
            ChannelStatisticsParameterJSON(['FSC-H', 'FSC-W'], ChannelStatisticType.MEDIAN)])
        stat_rec_id_list = flow_man.compute_statistics(input_json)
        self.assertIsNotNone(stat_rec_id_list)
        self.assertTrue(2 == len(stat_rec_id_list))

    def test_flowjo_workspace(self):
        workspace_file_name = "8_color_ICS.wsp"
        sample_file_name_list = ["101_DEN084Y5_15_E01_008_clean.fcs", "101_DEN084Y5_15_E03_009_clean.fcs",
                                 "101_DEN084Y5_15_E05_010_clean.fcs"]
        file_path = data_dir + "/" + workspace_file_name
        workspace_record_id: int
        with open(file_path, "rb") as f:
            file_bytes = f.read()
            workspace_record_id = flow_man.create_flowjo_workspace(
                FlowJoWorkspaceInputJson(workspace_file_name, file_bytes))
        workspace_record: DataRecord = data_record_manager.query_system_for_record(FlowJoWorkspaceRootModel.DATA_TYPE_NAME,
                                                                                   workspace_record_id)
        self.assertIsNotNone(workspace_record)
        for sample_file_name in sample_file_name_list:
            response_record_id: int
            file_path = data_dir + "/" + sample_file_name
            fcs_attachment_rec = _create_attachment_record_of_content(file_path)
            response_record_id = flow_man.upload_fcs_for_sample(UploadFCSInputJson(
                workspace_record.data_type_name, workspace_record.record_id, sample_file_name,
                fcs_attachment_rec.data_type_name, fcs_attachment_rec.record_id))
            self.assertTrue(response_record_id == workspace_record_id)



from __future__ import annotations

import time
from datetime import datetime
from typing import Any

import pytz

from sapiopycommons.general.exceptions import SapioException

__timezone = None
"""The default timezone. Use TimeUtil.set_default_timezone in a global context before making use of TimeUtil."""


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class TimeUtil:
    """
    A class that contains various date/time related utility methods. All times are based off of the timezone from
    the timezone() function. The default timezone is set by the timezone variable above this class definition.
    Since this is a global variable, every endpoint from the same server instance will use it. If you need to change
    the timezone, then call TimeUtil.set_default_timezone() somewhere in a global context in your server. A list of
    valid timezones can be found at https://en.wikipedia.org/wiki/List_of_tz_database_time_zones.

    If a timezone that is different from the default is needed but you don't want to change the default, a timezone name
    or UTC offset in seconds may be provided to each function. A UTC offset can be found at
    context.user.session_additional_data.utc_offset_seconds.

    Note that static date fields display their time in UTC instead of in whatever the server's time is. So when dealing
    with static date fields, use "UTC" as your input timezone.
    """
    @staticmethod
    def get_default_timezone() -> Any:
        """
        Returns the timezone that TimeUtil is currently using as its default.
        """
        global __timezone
        return __timezone

    @staticmethod
    def set_default_timezone(new_timezone: str | int) -> None:
        """
        Set the timezone used by TimeUtil to something new.

        :param new_timezone: The timezone to set the default to. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        """
        global __timezone
        __timezone = TimeUtil.__to_tz(new_timezone)

    @staticmethod
    def __to_tz(timezone: str | int = None) -> Any:
        """
        :param timezone: Either the name of a timezone, a UTC offset in seconds, or None if the default should be used.
        :return: The timezone object to use for the given input. If the input is None, uses the default timezone.
        """
        if isinstance(timezone, str):
            # PR-46571: Convert timezones to a UTC offset and then use that offset as the timezone. This is necessary
            # because pytz may return timezones from strings in Local Mean Time instead of a timezone with a UTC offset.
            # LMT may be a few minutes off of the actual time in that timezone right now.
            # https://stackoverflow.com/questions/35462876
            offset: int = TimeUtil.__get_timezone_offset(timezone)
            # This function takes an offset in minutes, so divide the provided offset seconds by 60.
            return pytz.FixedOffset(offset // 60)
        if isinstance(timezone, int):
            return pytz.FixedOffset(timezone // 60)
        if timezone is None:
            return TimeUtil.get_default_timezone()
        raise SapioException(f"Unhandled timezone object of type {type(timezone)}: {timezone}")

    @staticmethod
    def __get_timezone_offset(timezone: str | int | None) -> int:
        """
        :param timezone: Either the name of a timezone, a UTC offset in seconds, or None if the default should be used.
        :return: The UTC offset in seconds of the provided timezone.
        """
        if isinstance(timezone, int):
            return timezone
        if isinstance(timezone, str):
            timezone = pytz.timezone(timezone)
        if timezone is None:
            timezone = TimeUtil.get_default_timezone()
        return int(datetime.now(timezone).utcoffset().total_seconds())

    @staticmethod
    def current_time(timezone: str | int = None) -> datetime:
        """
        The current time as a datetime object.

        :param timezone: The timezone to initialize the current time with. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        """
        tz = TimeUtil.__to_tz(timezone)
        return datetime.now(tz)

    @staticmethod
    def now_in_millis() -> int:
        """
        The current time in milliseconds since the epoch.
        """
        return round(time.time() * 1000)

    @staticmethod
    def now_in_format(time_format: str, timezone: str | int = None) -> str:
        """
        The current time in some date format.

        :param time_format: The format to display the current time in. Documentation for how the time formatting works
            can be found at https://docs.python.org/3.10/library/datetime.html#strftime-and-strptime-behavior
        :param timezone: The timezone to initialize the current time with. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        """
        return TimeUtil.current_time(timezone).strftime(time_format)

    @staticmethod
    def millis_to_format(millis: int, time_format: str, timezone: str | int = None) -> str | None:
        """
        Convert the input time in milliseconds to the provided format. If None is passed to the millis parameter,
        None will be returned

        :param millis: The time in milliseconds to convert from.
        :param time_format: The format to display the input time in. Documentation for how the time formatting works
            can be found at https://docs.python.org/3.10/library/datetime.html#strftime-and-strptime-behavior
        :param timezone: The timezone to initialize the current time with. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        """
        if millis is None:
            return None

        tz = TimeUtil.__to_tz(timezone)
        return datetime.fromtimestamp(millis / 1000, tz).strftime(time_format)

    @staticmethod
    def format_to_millis(time_point: str, time_format: str, timezone: str | int = None) -> int:
        """
        Convert the input time from the provided format to milliseconds.

        :param time_point: The time in some date/time format to convert from.
        :param time_format: The format that the time_point is in. Documentation for how the time formatting works
            can be found at https://docs.python.org/3.10/library/datetime.html#strftime-and-strptime-behavior
        :param timezone: The timezone to initialize the current time with. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        """
        tz = TimeUtil.__to_tz(timezone)
        return int(datetime.strptime(time_point, time_format).replace(tzinfo=tz).timestamp() * 1000)

    # FR-47296: Provide functions for shifting between timezones.
    @staticmethod
    def shift_now(to_timezone: str = "UTC", from_timezone: str | None = None) -> int:
        """
        Take the current time in from_timezone and output the epoch timestamp that would display that same time in
        to_timezone. A use case for this is when dealing with static date fields to convert a provided timestamp to the
        value necessary to display that timestamp in the same way when viewed in the static date field.

        :param to_timezone: The timezone to shift to. If not provided, uses UTC.
        :param from_timezone: The timezone to shift from. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        :return: The epoch timestamp that would display as the same time in to_timezone as the current time in
            from_timezone.
        """
        millis: int = TimeUtil.now_in_millis()
        return TimeUtil.shift_millis(millis, to_timezone, from_timezone)

    @staticmethod
    def shift_millis(millis: int, to_timezone: str = "UTC", from_timezone: str | None = None) -> int:
        """
        Take a number of milliseconds for a time in from_timezone and output the epoch timestamp that would display that
        same time in to_timezone. A use case for this is when dealing with static date fields to convert a provided
        timestamp to the value necessary to display that timestamp in the same way when viewed in the static date field.

        :param millis: The time in milliseconds to convert from.
        :param to_timezone: The timezone to shift to. If not provided, uses UTC.
        :param from_timezone: The timezone to shift from. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        :return: The epoch timestamp that would display as the same time in to_timezone as the given time in
            from_timezone.
        """
        to_offset: int = TimeUtil.__get_timezone_offset(to_timezone) * 1000
        from_offset: int = TimeUtil.__get_timezone_offset(from_timezone) * 1000
        return millis + from_offset - to_offset

    @staticmethod
    def shift_format(time_point: str, time_format: str, to_timezone: str = "UTC", from_timezone: str | None = None) \
            -> int:
        """
        Take a timestamp for a time in from_timezone and output the epoch timestamp that would display that same time
        in to_timezone. A use case for this is when dealing with static date fields to convert a provided timestamp to
        the value necessary to display that timestamp in the same way when viewed in the static date field.

        :param time_point: The time in some date/time format to convert from.
        :param time_format: The format that the time_point is in. Documentation for how the time formatting works
            can be found at https://docs.python.org/3.10/library/datetime.html#strftime-and-strptime-behavior
        :param to_timezone: The timezone to shift to. If not provided, uses UTC.
        :param from_timezone: The timezone to shift from. If no timezone is provided, uses the global
            timezone variable set by the TimeUtil. A list of valid timezones can be found at
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones. May also accept a UTC offset in seconds.
        :return: The epoch timestamp that would display as the same time in to_timezone as the given time in
            from_timezone.
        """
        millis: int = TimeUtil.format_to_millis(time_point, time_format, from_timezone)
        return TimeUtil.shift_millis(millis, to_timezone, from_timezone)

    # FR-46154: Create a function that determines if a string matches a time format.
    @staticmethod
    def str_matches_format(time_point: str, time_format: str) -> bool:
        """
        Determine if the given string is recognized as a valid time in the given format.

        :param time_point: The time in some date/time format to check.
        :param time_format: The format that the time_point should be in. Documentation for how the time formatting works
            can be found at https://docs.python.org/3.10/library/datetime.html#strftime-and-strptime-behavior
        """
        try:
            # If this function successfully runs, then the time_point matches the time_format.
            TimeUtil.format_to_millis(time_point, time_format, "UTC")
            return True
        except Exception:
            return False



import unittest

from sapiopylib.rest.User import SapioUser

from sapiopycommons.general.accession_service import AccessionService
from data_type_models import SampleModel

user = SapioUser(url="https://linux-vm:8443/webservice/api", verify_ssl_cert=False,
                 guid="3c232543-f407-4828-aae5-b33d4cd31fa7",
                 username="yqiao_api", password="Password1!")

accession_service = AccessionService(user)

# This test needs to be run together with accessioning happening separately inside Sapio for creating samples/plates in 3D plater.
# Set plate to auto accession mode and ensure plate ID is non-unique to trigger global accessioning logic.
class AccessionServiceTest(unittest.TestCase):

    def test_sample_accession_with_config(self):
        """
        This is how platform accessions new samples that isn't an aliquot or pool.
        """
        id_list = accession_service.accession_with_config(SampleModel.DATA_TYPE_NAME, SampleModel.SAMPLEID__FIELD_NAME.field_name, 5)
        print("New Sample IDs: " + str(id_list) +
              ". Accession some more samples in ELN grabber and run this again and check continuity.")
        self.assertEqual(len(id_list), 5)

    def test_plate_id_nonunique_accession(self):
        """
        This will test the platform accessioning plate IDs.
        """
        # This is from MultiLayerPlatingConfigManager
        id_list = accession_service.get_global_affixed_id_in_batch(5, "", "", 4, 1000)
        print("New Plate IDs: " + str(id_list) + ". Check continuty in ELN using 3d plating auto accession plate ID, then make sure the plate ID is non-unique.")
        self.assertEqual(len(id_list), 5)

    def test_accession_pool_ids(self):
        # This is from CreateBoolBase, and MultiLayerPlatingRecordManager
        id_list = accession_service.get_affixed_id_in_batch(SampleModel.DATA_TYPE_NAME, SampleModel.SAMPLEID__FIELD_NAME.field_name,
                                                            10, "Pool-", "", None, 1000)
        print("New Pool IDs: " + str(id_list))
        self.assertEqual(len(id_list), 10)


import io

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.general.aliases import AliasUtil, SapioRecord, UserIdentifier
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.recordmodel.record_handler import RecordHandler


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class AttachmentUtil:
    @staticmethod
    def get_attachment_bytes(context: UserIdentifier, attachment: SapioRecord) -> bytes:
        """
        Get the data bytes for the given attachment record. Makes a webservice call to retrieve the data.

        :param context: The current webhook context or a user object to send requests from.
        :param attachment: The attachment record.
        :return: The bytes for the attachment's file data.
        """
        attachment = AliasUtil.to_data_record(attachment)
        dr_man = DataMgmtServer.get_data_record_manager(AliasUtil.to_sapio_user(context))
        with io.BytesIO() as data_sink:
            def consume_data(chunk: bytes):
                data_sink.write(chunk)
            dr_man.get_attachment_data(attachment, consume_data)
            data_sink.flush()
            data_sink.seek(0)
            file_bytes = data_sink.read()
        return file_bytes

    @staticmethod
    def set_attachment_bytes(context: UserIdentifier, attachment: SapioRecord,
                             file_name: str, file_bytes: bytes) -> None:
        """
        Set the attachment data for a given attachment record. Makes a webservice call to set the data.

        :param context: The current webhook context or a user object to send requests from.
        :param attachment: The attachment record. Must be an existing data record that is an attachment type.
        :param file_name: The name of the attachment.
        :param file_bytes: The bytes of the attachment data.
        """
        if attachment.record_id < 0:
            raise SapioException("Provided record cannot have its attachment data set, as it does not exist in the "
                                 "system yet.")
        attachment = AliasUtil.to_data_record(attachment)
        dr_man = DataMgmtServer.get_data_record_manager(AliasUtil.to_sapio_user(context))
        with io.BytesIO(file_bytes) as stream:
            dr_man.set_attachment_data(attachment, file_name, stream)

    @staticmethod
    def create_attachment(context: UserIdentifier, file_name: str, file_bytes: bytes,
                          wrapper_type: type[WrappedType]) -> WrappedType:
        """
        Create an attachment data type and initialize its attachment bytes at the same time.
        Makes a webservice call to create the attachment record and a second to set its bytes.

        :param context: The current webhook context or a user object to send requests from.
        :param file_name: The name of the attachment.
        :param file_bytes: THe bytes of the attachment data.
        :param wrapper_type: The attachment type to create.
        :return: A record model for the newly created attachment.
        """
        attachment: WrappedType = RecordHandler(context).create_models(wrapper_type, 1)[0]
        AttachmentUtil.set_attachment_bytes(context, attachment, file_name, file_bytes)
        return attachment



class StorageUtil:
    """
    A collection of utilities intended for converting to and from various forms of representing positions on a storage
    unit, such as character + integer (e.g. A1), two integers (e.g. (0, 0)), or a single integer index (e.g. 1).
    Integers are also sometimes zero-indexed and sometimes one-indexed, so both are supported.
    """
    @staticmethod
    def map_index_to_coordinate(index: int, size: int, fill_by_row: bool = True, zero_indexed_input: bool = False,
                                zero_indexed_output: bool = False, char_row_output: bool = True) \
            -> tuple[int | str, int]:
        """
        Convert an index representing a position on a plate/storage unit to a coordinate pair on that plate/storage
        unit, able to be used in row/column storage fields on records. This will output a value for any input index;
        it is up to the caller to determine if the output will actually fit on the plate/storage unit.
        (The index should be within the range [1 < index < rows * columns] for one-indexed values.)

        By default, expects the input to be a one-indexed integer filled row-by-row with the output being a
        character, integer pair where the output integer is one-indexed.

        :param index: The index to map to a coordinate position.
        :param size: The number of columns or rows in the plate/storage unit, depending on whether you are filling by
            row or by column.
        :param fill_by_row: If true, map positions row-by-row (A1, A2, A3... B1, B2...) and use the above size as the
            number of columns in the plate/storage unit. If false, map positions column-by-column (A1, B1, C1... A2,
            B2...) and use the above size as the number of rows.
        :param zero_indexed_input: If true, the input index is zero-indexed. If false, then they are one-indexed.
            This does not influence the output, only the function's understanding of the input.
        :param zero_indexed_output: If true, the output index is zero-indexed. If false, then it is one-indexed.
            Has no effect on the column output if the column is set to output as a character.
        :param char_row_output: If true, the output row value is converted to a character where 0 = A, 1 = B, 25 = Z,
            26 = AA, 27 = AB, etc. If false, then it is returned as an integer.
        :return: A tuple representing a coordinate pair (row value, column value). The row value may be either an
            integer or a string, while the column value is always an integer, influenced by the input parameters.
        """
        # If the given index isn't zero-indexed, then make it zero-indexed by subtracting one.
        if not zero_indexed_input:
            index -= 1

        row: int = index // size
        col: int = index % size

        # If fill by row is false, then the above calculations are flipped,
        # meaning the row is actually the column and vice versa.
        if not fill_by_row:
            temp = row
            row = col
            col = temp
        # The column and row are zero-indexed by default. If it should be one-indexed, add one.
        if not zero_indexed_output:
            col += 1
            # Only add one to the row if it won't be converted to a character.
            if not char_row_output:
                row += 1
        return StorageUtil.map_index_to_char(row, True) if char_row_output else row, col

    @staticmethod
    def map_coordinate_to_index(row: int | str, col: int | str, size: int, fill_by_row: bool = True,
                                zero_indexed_input: bool = False, zero_indexed_output: bool = False) -> int:
        """
        Map row and column coordinates on a plate/storage unit to the index of that position.

        By default, expects the input to be provided as a character, integer pair and outputs a single row-by-row
        one-indexed integer.

        :param row: The row coordinate of the position as a string of characters from A-Z or a zero-indexed integer.
        :param col: The column coordinate of the position as an integer which may be zero-indexed or one-indexed.
            (This integer may be in string form, such as is the case with column fields on storable records.)
        :param size: The number of columns or rows in the plate/storage unit, depending on whether you are filling by
            row or by column.
        :param fill_by_row: If true, map positions row-by-row (A1, A2, A3... B1, B2...) and use the above size as the
            number of columns in the plate/storage unit. If false, map positions column-by-column (A1, B1, C1... A2,
            B2...) and use the above size as the number of rows.
        :param zero_indexed_input: If true, the input coordinates for the row and column is zero-indexed. If false,
            then they are one-indexed. This does not influence the output, only the function's understanding of the
            input. This also has no effect if the input row is a string.
        :param zero_indexed_output: If true, the output index is zero-indexed. If false, then it is one-indexed.
        :return: The index of the storage position at the input row and column.
        """
        # If the column was provided as a string, cast it to an int.
        if isinstance(col, str):
            col: int = int(col)
        # If the input isn't zero-indexed, then make it zero-indexed.
        if not zero_indexed_input:
            col -= 1
            # Only subtract from the row if it's already in integer form.
            # If it's a string, it'll be converted to a zero-indexed integer.
            if isinstance(row, int):
                row -= 1
        # If the input row is a string, convert it to a zero-indexed integer.
        if isinstance(row, str):
            row: int = StorageUtil.map_char_to_index(row, True)

        # Convert the row and column indices to a singular index across the entire storage unit.
        index: int = row * size + col if fill_by_row else col * size + row

        # The index is zero-indexed by default. If it should be one-indexed, add one.
        if not zero_indexed_output:
            index += 1
        return index

    @staticmethod
    def map_index_to_char(index: int, zero_indexed_input: bool = False) -> str:
        """
        Map a given base-10 integer to a base-26 value where 0 = A, 1 = B, 25 = Z, 26 = AA, 27 = AB, etc.
        Useful for mapping the index of a row to the character(s) representing that row in a storage unit.
        May also be used for mapping the index to an Excel sheet's columns.

        By default, expects the input as a one-indexed value.

        :param index: The index to map to a character.
        :param zero_indexed_input: If true, the input index is zero-indexed. If false, then they are one-indexed.
            This does not influence the output, only the function's understanding of the input.
        :return: The input integer mapped to a string representing that integer's position.
        """
        # If the given index isn't zero-indexed, then make it zero-indexed by subtracting one.
        if not zero_indexed_input:
            index -= 1
        chars: str = ""
        while index >= 0:
            # Add new characters to the front of the string.
            chars = chr(ord("A") + index % 26) + chars
            # Reduce the index by the amount accounted for by the character that was just added.
            index = index // 26 - 1
        return chars

    @staticmethod
    def map_char_to_index(chars: str, zero_indexed_output: bool = False) -> int:
        """
        Map a given base-26 value of characters to a base-10 integer where A = 0, B = 1, Z = 25, AA = 26, AB = 27, etc.
        Useful for mapping the character(s) representing a row in a storage unit to that row's index.
        May also be used for mapping the index of an Excel sheet's columns.

        By default, provides the output as a one-indexed value.

        :param chars: A string of characters to be converted to an index. Characters are expected to be uppercase
            characters in the range A to Z.
        :param zero_indexed_output: If true, the output index is zero-indexed. If false, then it is one-indexed.
        :return: The input character(s) converted to an index.
        """
        # Reverse iterate over the characters of the string and determine the value of each individual character.
        # The value is multiplied by the base of the character given its digit position (26^0, 26^1, etc.)
        value: int = 0
        for i, c in enumerate(reversed(chars)):
            value += (ord(c) - ord("A") + 1) * (26 ** i)
        # The character value is one-indexed by default. If it should be zero-indexed, subtract one.
        if zero_indexed_output:
            value -= 1
        return value



# Add facilities to test chemistry endpoints. This will run its own webserver.
# These testing will require working with a server with chemistry license and is set up with chemistry correctly.
from io import BytesIO

from sapiopylib.rest.ClientCallbackService import ClientCallback
from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.DataTypeService import DataTypeManager
from sapiopylib.rest.WebhookService import AbstractWebhookHandler, WebhookConfiguration, WebhookServerFactory
from sapiopylib.rest.pojo.CustomReport import RawReportTerm, RawTermOperation
from sapiopylib.rest.pojo.datatype.FieldDefinition import VeloxDoubleFieldDefinition
from sapiopylib.rest.pojo.webhook.ClientCallbackRequest import *
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.pojo.webhook.WebhookResult import SapioWebhookResult
from sapiopylib.rest.utils.autopaging import QueryDataRecordByIdListAutoPager, QueryAllRecordsOfTypeAutoPager

from data_type_models import *
from sapiopycommons.multimodal.multimodal import MultiModalManager
from sapiopycommons.multimodal.multimodal_data import *


def __select_type_and_item_to_load(client_callback: ClientCallback) -> tuple[str, bool]:
    options = ["Compound Parts", "Compound Samples", "Chem Reagent Parts", "Chem Reagent Samples"]
    selected = client_callback.show_option_dialog(OptionDialogRequest(
        "Select Item Type", "Select one of the following item types that correspond to what you are registering",
        options))
    data_type: str
    is_adding_items: bool
    if selected == 0:
        data_type = CompoundPartModel.DATA_TYPE_NAME
        is_adding_items = False
    elif selected == 1:
        data_type = CompoundPartModel.DATA_TYPE_NAME
        is_adding_items = True
    elif selected == 2:
        data_type = ChemicalReagentPartModel.DATA_TYPE_NAME
        is_adding_items = False
    elif selected == 3:
        data_type = ChemicalReagentPartModel.DATA_TYPE_NAME
        is_adding_items = True
    else:
        raise ValueError("Invalid selection index: " + str(selected))
    return data_type, is_adding_items


def download_chem_image(context: SapioWebhookContext) -> SapioWebhookResult:
    """
    Upload a chemical smiles user enters and download its SVG image.
    """
    user = context.user
    client_callback = DataMgmtServer.get_client_callback(user)
    smiles = client_callback.show_input_dialog(InputDialogCriteria("Enter a structure SMILES or RxN to load", ""))
    if not smiles:
        return SapioWebhookResult(False)
    reg_man = MultiModalManager(user)
    is_reaction_sel = client_callback.show_option_dialog(OptionDialogRequest("Is this a reaction?", "", ["Yes", "No"]))
    is_reaction = is_reaction_sel == 0
    image_list = reg_man.load_image_data(ImageDataRequestPojo([smiles], is_reaction))
    image_svg_text = image_list[0]
    with BytesIO(image_svg_text.encode('utf-8')) as io:
        client_callback.send_file("image.svg", False, io)
    return SapioWebhookResult(True)


def register_non_interactively(context: SapioWebhookContext) -> SapioWebhookResult:
    """
    Without asking user for anything in between the registration,
    complete registration using non-interactive endpoints only.

    The only interactions are those provided within this webhook explicitly or custom rules/on-save actions.
    """
    user = context.user
    client_callback = DataMgmtServer.get_client_callback(user)
    data_type, is_adding_items = __select_type_and_item_to_load(client_callback)

    smiles = client_callback.show_input_dialog(InputDialogCriteria("Enter SMILES to register", ""))
    if not smiles:
        return SapioWebhookResult(False)
    reg_man = MultiModalManager(user)
    loaded_result: PyMoleculeLoaderResult = reg_man.load_compounds(CompoundLoadRequestPojo(
        data_type, ChemLoadType.SMILES_LIST, data_list=[smiles]))
    if loaded_result.errorList:
        client_callback.display_warning(
            "There is an error loading the structure. Registration was aborted.: " + str(loaded_result.errorList[0]))
    complete_result: ChemCompleteImportPojo = reg_man.register_compounds(
        ChemRegisterRequestPojo(data_type, loaded_result.compoundList))
    client_callback.display_info(
        "Registered " + str(len(complete_result.newPartList)) + " new parts and used " +
        str(complete_result.numOldParts) + " old parts in this registration.")
    return SapioWebhookResult(True)


def register_reaction(context: SapioWebhookContext) -> SapioWebhookResult:
    user = context.user
    client_callback = DataMgmtServer.get_client_callback(user)

    rxn = client_callback.show_input_dialog(InputDialogCriteria("Enter RxN to register", ""))
    if not rxn:
        return SapioWebhookResult(False)
    reg_man = MultiModalManager(user)
    loaded_result = reg_man.load_reactions(rxn)
    registered = reg_man.register_reactions(loaded_result.reactionRxn)
    client_callback.display_info("Successfully registered with Record: " + str(registered.fields))
    return SapioWebhookResult(True)


def structure_search(context: SapioWebhookContext) -> SapioWebhookResult:
    """
    Perform a substructure search or a similarity search.
    """
    user = context.user
    client_callback = DataMgmtServer.get_client_callback(user)
    option = client_callback.show_option_dialog(OptionDialogRequest(
        "Select Search Type", "",
        ["Compound Substructure", "Compound Similarity", "Reaction Substructure"]))
    search_type: ChemSearchType
    if option == 0:
        search_type = ChemSearchType.COMPOUND_SUBSTRUCTURE
    elif option == 1:
        search_type = ChemSearchType.COMPOUND_SIMILARITY
    else:
        search_type = ChemSearchType.REACTION_SUBSTRUCTURE
    query: str = client_callback.show_input_dialog(InputDialogCriteria(
        "Enter Search Text",
        "Enter the SMARTS for compound substructure search or SMILES for compound similarity search, "
        "or rSMARTS for reaction substructure search."))

    join_sapio_type: str
    if search_type is ChemSearchType.COMPOUND_SUBSTRUCTURE or search_type is ChemSearchType.COMPOUND_SIMILARITY:
        option = client_callback.show_option_dialog(OptionDialogRequest(
            "Select Join Type", "Select a Sapio registry type you are joining result with.",
            ["Compounds", "Chemical Reagants", "HELM Chemical Properties"]))
        if option == 0:
            join_sapio_type = CompoundPartModel.DATA_TYPE_NAME
        elif option == 1:
            join_sapio_type = ChemicalReagentPartModel.DATA_TYPE_NAME
        else:
            join_sapio_type = HelmChemPropertyModel.DATA_TYPE_NAME
    else:
        join_sapio_type = ReactionModel.DATA_TYPE_NAME
    sim_search_upper: float | None = None
    if search_type is ChemSearchType.COMPOUND_SIMILARITY:
        field_def = VeloxDoubleFieldDefinition("Upper", "Upper", "Upper",
                                               0.0, 1.0, 0.95, 5)
        field_def.editable = True
        sim_search_upper = client_callback.show_input_dialog(InputDialogCriteria(
            "Enter Similarity Upper Bound", "", field_def))

    next_page_context: ChemQuickSearchContextData | None = None

    if not query:
        return SapioWebhookResult(False)
    reg_man = MultiModalManager(user)
    dt_man = DataTypeManager(user)
    temp_dt: TemporaryDataType = dt_man.get_temporary_data_type(join_sapio_type)
    while True:
        search_result = reg_man.search_structures(
            ChemSearchRequestPojo(query, search_type, join_sapio_type, next_page_context, sim_search_upper))
        if search_result.recordIdListOfPage:
            records: list[DataRecord] = QueryDataRecordByIdListAutoPager(
                temp_dt.data_type_name, search_result.recordIdListOfPage, user).get_all_at_once()
            client_callback.show_table_entry_dialog(TableEntryDialogRequest(
                "Current Page Results", "", temp_dt, [x.fields for x in records]))
        else:
            client_callback.display_popup(DisplayPopupRequest(
                "No Results in Page", "This page may not contain any results. Good luck next time!",
                PopupType.Info))
        if not search_result.nextPageAvailable:
            client_callback.display_info("Hit End of Results. No further pages available.")
            return SapioWebhookResult(True)
        next_page_context = search_result.nextPageContext

        selected = client_callback.show_option_dialog(OptionDialogRequest("Continue?", "Continue to next page?", ["Yes", "No"]))
        if selected is None or selected != 0:
            return SapioWebhookResult(True)


def export_sdf(context: SapioWebhookContext) -> SapioWebhookResult:
    user = context.user
    custom_report_manager = DataMgmtServer.get_custom_report_manager(user)
    client_callback = DataMgmtServer.get_client_callback(user)
    report = custom_report_manager.run_quick_report(RawReportTerm(ChemicalReagentPartModel.DATA_TYPE_NAME, "RecordId", RawTermOperation.GREATER_THAN_OPERATOR, "0"), 100, 0)
    result_table: list[list[Any]] = report.result_table
    record_id_col_index: int = [idx for idx, x in enumerate(report.column_list) if x.data_field_name == "RecordId"][0]
    record_id_list = [int(x[record_id_col_index]) for x in result_table if x[record_id_col_index] is not None]
    if not record_id_list:
        client_callback.display_warning("There are no Chemical Reagent Part to test with. Please register at least 1 such record and test again.")
        return SapioWebhookResult(True)
    custom_payload: dict[int, dict[str, str]] = dict()
    for i, record_id in enumerate(record_id_list):
        custom_payload[record_id] = {"pyIndex": str(i)}
    request = ChemExportSDFRequest(ChemicalReagentPartModel.DATA_TYPE_NAME, record_id_list,
                                   additionalPropertiesByRecordId=custom_payload)
    sdf_text = MultiModalManager(user).export_to_sdf(request)
    with BytesIO(sdf_text.encode()) as io:
        client_callback.send_file("output.sdf", False, io)
    return SapioWebhookResult(True)


class ChemistryRegistrationTestHandler(AbstractWebhookHandler):
    """
    This one will test the facilities to parse a chemical structure and return a svg file to user's browser.
    """

    def run(self, context: SapioWebhookContext) -> SapioWebhookResult:
        options = ['Download Image', 'Register Non-Interactively', 'Structure Search',
                   'Register Reaction', "Export SDF"]
        user = context.user
        client_callback = DataMgmtServer.get_client_callback(user)
        selected = client_callback.show_option_dialog(OptionDialogRequest(
            "Select Test Option", "Select one of the test options below", options, closable=True))
        if selected is None:
            return SapioWebhookResult(False)
        if selected == 0:
            return download_chem_image(context)
        elif selected == 1:
            return register_non_interactively(context)
        elif selected == 2:
            return structure_search(context)
        elif selected == 3:
            return register_reaction(context)
        elif selected == 4:
            return export_sdf(context)
        return SapioWebhookResult(False)

config: WebhookConfiguration = WebhookConfiguration(verify_sapio_cert=False, debug=True)
config.register('/chem_test', ChemistryRegistrationTestHandler)

app = WebhookServerFactory.configure_flask_app(app=None, config=config)
app.run(host="0.0.0.0", port=8099)



import io
from enum import Enum

import paramiko
from paramiko import pkey
from paramiko.sftp_client import SFTPClient

from sapiopycommons.general.exceptions import SapioException


class SFTPAuthMethod(Enum):
    """
    An enum being used to specify connection type to the target server.
    """
    PASSWORD = 0
    """Connection is being done via Password."""
    FILEPATH = 1
    """Connection is being done using a private key file in the codebase."""
    KEY_STRING = 2
    """Connection is being done using a private key in string form."""


class SFTPBuilder:
    """
    A class for making SFTP connections.
    """

    @staticmethod
    def open_sftp(username: str, host: str, port: int, authentication: str,
                  connection_type: SFTPAuthMethod = SFTPAuthMethod.PASSWORD) -> SFTPClient:
        """
        Builds a SFTP client from user input.

        :param username: The username of the individual trying to connect to the target server.
        :param host: The hostname/IP address of the target server.
        :param port: The port number used to connect to the target server.
        :param authentication: The string used to connect to the target server. This could hold a filepath, a password
            or a private key in string form depending on the connection_type parameter.

            If authentication is a private key string, they are generally formated like this:
            -----BEGIN OPENSSH PRIVATE KEY-----\n
            asdfh;hjadfh;jghajdg54646+5fasdfadlajklgajd'gj'ajg654564\n
            asdkjfhj;kghj;ahj;wh41234hjadjkhhdsgadshjkdghjshdlsds468\n
            ....

        :param connection_type: This enum is used to specify how the connection to the target server is being made.
            The options are:
            (0) PASSWORD: This means that the authentication parameter contains a password that will be used to connect to the server
            (1) FILEPATH: This means that the authentication parameter contains a filepath leading to a private key file stored in the codebase
            (2) KEY_STRING: This means that the authentication parameter contains the private key in string form

        """

        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy)

        if connection_type == SFTPAuthMethod.FILEPATH:
            client.connect(host, username=username, port=port, key_filename=authentication)
            return client.open_sftp()

        if connection_type == SFTPAuthMethod.KEY_STRING:
            private_key: pkey = paramiko.RSAKey.from_private_key(io.StringIO(authentication))
            client.connect(host, username=username, port=port, pkey=private_key)
            return client.open_sftp()

        if connection_type == SFTPAuthMethod.PASSWORD:
            client.connect(host, username=username, password=authentication, port=port)
            return client.open_sftp()

        raise SapioException("The SFTPAuthMethod enumerator was not properly specified.")



import unittest

from sapiopycommons.multimodal.multimodal import MultiModalManager
from sapiopycommons.multimodal.multimodal_data import *
from data_type_models import *

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.User import SapioUser

user = SapioUser(url="https://linux-vm:8443/webservice/api", verify_ssl_cert=False,
                 guid="ae4e92e1-34c6-4336-8d80-7f6804d3e51c",
                 username="yqiao_api", password="Password1!")

multi_modal_manager = MultiModalManager(user)


class ChemCurationTest(unittest.TestCase):
    def test_curation_reg(self):
        with open("curation_queue_test.sdf", "rb") as f:
            file_data: bytes = f.read()
            loader_result = multi_modal_manager.load_compounds(
                CompoundLoadRequestPojo(ChemicalReagentPartModel.DATA_TYPE_NAME, ChemLoadType.SDF_FILE,
                                        file_data=file_data))
            compound_list = loader_result.compoundList
            for compound in compound_list:
                self.assertTrue(compound.originalMol.CXSMILESHash)
                self.assertTrue(compound.originalMol.registrationHash)
            multi_modal_manager.register_compounds(ChemRegisterRequestPojo(
                ChemicalReagentPartModel.DATA_TYPE_NAME, compound_list))
            # User manually log in and resolve the curation queue.
            print("Loaded! Now play with curation queue inside Sapio!")



from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext

from sapiopycommons.general.aliases import RecordIdentifier, ExperimentIdentifier, AliasUtil, DataTypeIdentifier
from sapiopycommons.general.exceptions import SapioException


class SapioNavigationLinker:
    """
    Given a URL to a system's webservice API (example: https://company.exemplareln.com/webservice/api), construct
    URLs for navigation links to various locations in the system.
    """
    base_url: str

    def __init__(self, url: str | SapioUser | SapioWebhookContext):
        """
        :param url: A user or context object that is being used to send requests to a Sapio system, or a URL to a
            system's webservice API.
        """
        if isinstance(url, SapioWebhookContext):
            url = url.user.url
        elif isinstance(url, SapioUser):
            url = url.url
        self.base_url = url.rstrip("/").replace('webservice/api', 'veloxClient')

    def data_record(self, record_identifier: RecordIdentifier, data_type_name: DataTypeIdentifier | None = None) -> str:
        """
        :param record_identifier: An object that can be used to identify a record in the system, be that a record ID,
            a data record, or a record model.
        :param data_type_name: If the provided record identifier is a record ID, then the data type name of the record
            must be provided in this parameter. Otherwise, this parameter is ignored.
        :return: A URL for navigating to the input record.
        """
        record_id: int = AliasUtil.to_record_id(record_identifier)
        if data_type_name:
            data_type_name = AliasUtil.to_data_type_name(data_type_name)
        if not isinstance(record_identifier, int):
            data_type_name = AliasUtil.to_data_type_name(record_identifier)
        if not data_type_name:
            raise SapioException("Unable to create a data record link without a data type name. "
                                 "Only a record ID was provided.")
        return self.base_url + f"/#dataType={data_type_name};recordId={record_id};view=dataRecord"

    def experiment(self, experiment: ExperimentIdentifier) -> str:
        """
        :param experiment: An object that can be used to identify an experiment in the system, be that an experiment
            object, experiment protocol, or a notebook ID.
        :return: A URL for navigating to the input experiment.
        """
        return self.base_url + f"/#notebookExperimentId={AliasUtil.to_notebook_id(experiment)};view=eln"



from __future__ import annotations

from weakref import WeakValueDictionary

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.DataRecord import DataRecord
from sapiopylib.rest.pojo.eln.SapioELNEnums import ElnBaseDataType
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager, RecordModelInstanceManager
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.general.aliases import FieldMap, DataTypeIdentifier, AliasUtil
from sapiopycommons.general.exceptions import SapioException


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class OnSaveRuleHandler:
    """
    A class which helps with the parsing and navigation of the on save rule result map of a webhook context.
    """
    __context: SapioWebhookContext
    """The context that this handler is working from."""

    __inst_man: RecordModelInstanceManager
    """The record model instance manager, used for wrapping the data records as record models."""

    # Reformatted and cached version of the Velox on save rule result map for easier handling.
    __records: dict[str, set[DataRecord]]
    """A mapping of data type to the set of data records from the context that match that data type."""
    __base_id_to_records: dict[int, dict[str, set[DataRecord]]]
    """A mapping of record IDs of records in the context.data_record_list to the sets of data records related to that
    record, each set of records being mapped by its data type."""
    __field_maps: dict[str, dict[int, FieldMap]]
    """A mapping of data type to the field maps from the context that match that data type. In order to prevent
    duplicate field maps, each field map is in a dict keyed by the RecordId field in the field map, since field maps
    are just dictionaries and dictionaries aren't hashable and therefore can't go in a set."""
    __base_id_to_field_maps: dict[int, dict[str, dict[int, FieldMap]]]
    """A mapping of record IDs of records in the context.data_record_list to the field maps related to that
    record, each grouping of field maps being mapped by its data type."""

    __instances: WeakValueDictionary[SapioUser, OnSaveRuleHandler] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, context: SapioWebhookContext):
        if context.velox_on_save_result_map is None:
            raise SapioException("No Velox on save rule result map in context for OnSaveRuleHandler to parse.")
        user = context if isinstance(context, SapioUser) else context.user
        obj = cls.__instances.get(user)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[user] = obj
        return obj

    def __init__(self, context: SapioWebhookContext):
        if self.__initialized:
            return
        self.__initialized = True

        if context.velox_on_save_result_map is None:
            raise SapioException("No Velox on save rule result map in context for OnSaveRuleHandler to parse.")
        self.__context = context
        self.__inst_man = RecordModelManager(context.user).instance_manager
        self.__cache_records()

    def __cache_records(self) -> None:
        """
        Cache the records from the context into dictionaries. Two caches are created. One cache maps the data type of
        each record to a set of all records of that data type. The other cache maps the record ID that the records relate
        to with another dict that maps the data types to the records of that type. Doesn't cache any relationship info.
        """
        self.__records = {}
        self.__base_id_to_records = {}
        # Each record ID in the context has a list of results for that record.
        for record_id, rule_results in self.__context.velox_on_save_result_map.items():
            # Keep track of the records for this specific record ID.
            id_dict: dict[str, set[DataRecord]] = {}
            # The list of results for a record consist of a list of data records and a VeloxType that specifies
            # how the records in the list relate to the main record.
            for record_result in rule_results:
                # For the purposes of caching, we don't care about the VeloxType.
                for record in record_result.data_records:
                    # Get the data type of this record. If this is an ELN type, ignore the digits.
                    data_type: str = record.data_type_name
                    # PR-46331: Ensure that all ELN types are converted to their base data type name.
                    if ElnBaseDataType.is_eln_type(data_type):
                        data_type = ElnBaseDataType.get_base_type(data_type).data_type_name
                    # Update the list of records of this type that exist so far globally.
                    self.__records.setdefault(data_type, set()).add(record)
                    # Do the same for the list of records of this type that relate to this record ID.
                    id_dict.setdefault(data_type, set()).add(record)
            # Update the related records for this record ID.
            self.__base_id_to_records.update({record_id: id_dict})

        self.__field_maps = {}
        self.__base_id_to_field_maps = {}
        # Repeat the same thing for the field map results.
        for record_id, rule_results in self.__context.velox_on_save_field_map_result_map.items():
            id_dict: dict[str, dict[int, FieldMap]] = {}
            for record_result in rule_results:
                data_type: str = record_result.velox_type_pojo.data_type_name
                if ElnBaseDataType.is_eln_type(data_type):
                    data_type = ElnBaseDataType.get_base_type(data_type).data_type_name
                for field_map in record_result.field_map_list:
                    rec_id: int = field_map.get("RecordId")
                    self.__field_maps.setdefault(data_type, {}).update({rec_id: field_map})
                    id_dict.setdefault(data_type, {}).update({rec_id: field_map})
            self.__base_id_to_field_maps.update({record_id: id_dict})

    def get_base_record_ids(self) -> list[int]:
        """
        :return: A list of the record IDs that may be used with the get_records and get_models functions. These are the
            record IDs to the records that caused the rule to trigger.
        """
        return list(self.__base_id_to_records.keys())

    def get_field_maps_base_record_ids(self) -> list[int]:
        """
        :return: A list of the record IDs that may be used with the get_field_maps function. These are the
            record IDs to the records that caused the rule to trigger.
        """
        return list(self.__base_id_to_field_maps.keys())

    def get_records(self, data_type: DataTypeIdentifier, record_id: int | None = None) -> list[DataRecord]:
        """
        Get records from the cached context with the given data type. Capable of being filtered to searching within
        the context of a record ID. If the given data type or record ID does not exist in the context,
        returns an empty list.

        :param data_type: The data type of the records to return.
        :param record_id: The record ID of the base record to search from. If None, returns the records that match the
            data type from every ID. If an ID is provided, but it does not exist in the context, returns an empty list.
        :return: The records from the context that match the input parameters.
        """
        data_type: str = AliasUtil.to_data_type_name(data_type)
        records: dict[str, set[DataRecord]] = self.__base_id_to_records.get(record_id, {}) if record_id else self.__records
        return list(records.get(data_type, []))

    # FR-46701: Add functions to the rule handlers for accessing the field maps of inaccessible records in the context.
    def get_field_maps(self, data_type: DataTypeIdentifier, record_id: int | None = None) -> list[FieldMap]:
        """
        Get field maps from the cached context with the given data type. Capable of being filtered to searching within
        the context of a record ID. If the given data type or record ID does not exist in the context,
        returns an empty list.

        Field maps will only exist in the context if the data record that the fields are from is no longer accessible
        to the user. This can occur because the data record was deleted, or because the user does not have access to the
        record due to ACL.

        :param data_type: The data type of the field maps to return.
        :param record_id: The record ID of the base record to search from. If None, returns the field maps that match
            the data type from every ID. If an ID is provided, but it does not exist in the context, returns an empty
            list.
        :return: The field maps from the context that match the input parameters.
        """
        data_type: str = AliasUtil.to_data_type_name(data_type)
        field_maps: dict[str, dict[int, FieldMap]] = self.__base_id_to_field_maps.get(record_id, {}) if record_id else self.__field_maps
        return list(field_maps.get(data_type, {}).values())

    def get_models(self, wrapper_type: type[WrappedType], record_id: int | None = None) -> list[WrappedType]:
        """
        Get records from the cached context with the given data type. Capable of being filtered to searching within
        the context of a record ID. If the given data type or record ID does not exist in the context,
        returns an empty list.

        :param wrapper_type: The record model wrapper to use.
        :param record_id: The record ID of the base record to search from. If None, returns the records that match the
            data type from ID. If an ID is provided, but it does not exist in the context, returns an empty list.
        :return: The record models from the context that match the input parameters.
        """
        dt: str = wrapper_type.get_wrapper_data_type_name()
        return self.__inst_man.add_existing_records_of_type(self.get_records(dt, record_id), wrapper_type)
